from ..codegen.schema.reference import SchemaReference
import re


class Reference(SchemaReference):
    @property
    def as_qn_path(self):
        return "/".join(self.snakify().qualified_name.split(".")) + ".py"

    @property
    def as_type_path(self):
        return "/".join(self.snakify().type_name.split(".")) + ".py"

    @property
    def as_module_path(self):
        return "/".join(self.snakify().module_name.split(".")) + ".py"

    @property
    def as_qualified_ident(self) -> str:
        cameled = self.camelify()
        return cameled.module_name[0].upper() + cameled.module_name[1:] + cameled.type_name

    def camelify(self) -> "Reference":
        def camelify(name):
            return "".join(p[0].upper() + p[1:] for p in re.split("[._]", name) if p)

        return Reference(
            {
                "module_name": camelify(self.module_name),
                "qualified_name": camelify(self.qualified_name),
                "type_name": camelify(self.type_name),
            }
        )

    def snakify(self) -> "Reference":
        def snakify(name):
            first_pass = re.sub("([^.])([A-Z][a-z]+)", r"\1_\2", name)
            return re.sub("([a-z0-9])([A-Z])", r"\1_\2", first_pass).lower()

        return Reference(
            {
                "module_name": snakify(self.module_name),
                "qualified_name": snakify(self.qualified_name),
                "type_name": snakify(self.type_name),
            }
        )



#! /usr/bin/env python3

import os.path

from cached_property import cached_property
from typing import Callable, List, Dict, Any

from idol import scripter
from idol.cli import start, CliConfig
from idol.functional import OrderedObj, Alt
from idol.generator import (
    GeneratorParams,
    GeneratorConfig,
    build,
    GeneratorAcc,
    TypeDeconstructor,
    Path,
    get_material_type_deconstructor,
    TypeStructDeconstructor,
    ScalarDeconstructor,
    GeneratorContext,
    GeneratorFileContext,
    Exported,
    Expression,
    import_expr,
    get_safe_ident,
    get_tag_values,
    includes_tag,
    ExternFileContext,
    AbstractGeneratorFileContext,
)
from idol.py.schema.primitive_type import PrimitiveType
from idol.py.schema.reference import Reference
from idol.py.schema.type import Type

ExportedAny: Exported = Exported(Path("typing"), "Any")


class IdolPy(GeneratorContext):
    scaffolds: Dict[str, "IdolPyScaffoldFile"]
    codegens: Dict[str, "IdolPyCodegenFile"]
    scaffold_impl: "Callable[[IdolPy, Type, Path], IdolPyScaffoldFile]"
    codegen_impl: "Callable[[IdolPy, Type, Path], IdolPyCodegenFile]"

    def __init__(
        self,
        config: GeneratorConfig,
        scaffold_impl: "Callable[[IdolPy, Type, Path], IdolPyScaffoldFile]" = None,
        codegen_impl: "Callable[[IdolPy, Type, Path], IdolPyCodegenFile]" = None,
    ):
        super(IdolPy, self).__init__(GeneratorAcc(), config)
        self.scaffolds = {}
        self.codegens = {}
        self.scaffold_impl = scaffold_impl or IdolPyScaffoldFile
        self.codegen_impl = codegen_impl or IdolPyCodegenFile

    def scaffold_file(self, ref: Reference) -> "IdolPyScaffoldFile":
        path = self.state.reserve_path(**self.config.paths_of(scaffold=ref))
        t = self.config.params.all_types.obj[ref.qualified_name]

        if ref.qualified_name not in self.scaffolds:
            self.scaffolds[ref.qualified_name] = self.scaffold_impl(self, t, path)

        return self.scaffolds[ref.qualified_name]

    def codegen_file(self, ref: Reference) -> "IdolPyCodegenFile":
        path = self.state.reserve_path(**self.config.paths_of(codegen=ref))
        t = self.config.params.all_types.obj[ref.qualified_name]

        if ref.qualified_name not in self.codegens:
            self.codegens[ref.qualified_name] = self.codegen_impl(self, t, path)

        return self.codegens[ref.qualified_name]

    @cached_property
    def idol_py_file(self) -> "IdolPyFile":
        path = self.state.reserve_path(runtime=self.config.codegen_root + "/__idol__.py")
        return IdolPyFile(self, path)

    def render(self) -> OrderedObj[str]:
        for i, t in enumerate(self.config.params.scaffold_types.values()):
            for export in self.scaffold_file(t.named).declared_type_ident:
                print(
                    f"Rendered {export.ident} to {export.path.path} ({i} / {len(self.config.params.scaffold_types)})"
                )
                break
            else:
                print(
                    f"Skipped {t.named.as_qualified_ident} ({i} / {len(self.config.params.scaffold_types)})"
                )

        return self.state.render(
            dict(
                codegen=[
                    "DO NOT EDIT",
                    "This file was generated by idol_py, any changes will be lost when idol_py is rerun again",
                ],
                scaffold=[
                    "This file was scaffold by idol_py, but it will not be overwritten, so feel free to edit.",
                    "This file will be regenerated if you delete it.",
                ],
            )
        )


class IdolPyCodegenFile(GeneratorFileContext):
    t: Type
    idol_py: IdolPy
    t_decon: TypeDeconstructor

    def __init__(self, idol_py: IdolPy, t: Type, path: Path):
        self.t = t
        self.idol_py = idol_py
        self.t_decon = TypeDeconstructor(t)
        super(IdolPyCodegenFile, self).__init__(idol_py, path)

        self.reserve_ident(self.default_type_name)

    @cached_property
    def declared_type_ident(self) -> Alt[Exported]:
        type_ident: Alt[Exported] = Alt(
            v for typestruct in self.typestruct for v in typestruct.declared_ident
        )

        type_ident ^= Alt(v for enum in self.enum for v in enum.declared_ident)
        type_ident ^= Alt(v for struct in self.struct for v in struct.declared_ident)

        return type_ident

    @property
    def default_type_name(self) -> str:
        return self.t.named.as_qualified_ident

    @cached_property
    def typestruct(self) -> "Alt[IdolPyCodegenTypeStructDeclaration]":
        return Alt(
            IdolPyCodegenTypeStructDeclaration(self, ts_decon)
            for ts_decon in self.t_decon.get_typestruct()
        )

    @cached_property
    def enum(self) -> "Alt[IdolPyCodegenEnum]":
        return Alt(IdolPyCodegenEnum(self, options) for options in self.t_decon.get_enum())

    @cached_property
    def struct(self) -> "Alt[IdolPyCodegenStruct]":
        return Alt(
            IdolPyCodegenStruct(
                self,
                OrderedObj.from_iterable(
                    OrderedObj({k: IdolPyCodegenTypeStruct(self.idol_py, ts_decon)})
                    for k, ts_decon in fields
                ),
            )
            for fields in self.t_decon.get_struct()
        )


class IdolPyCodegenStruct(GeneratorFileContext):
    codegen_file: IdolPyCodegenFile
    fields: OrderedObj["IdolPyCodegenTypeStruct"]

    def __init__(
        self, codegen_file: IdolPyCodegenFile, fields: "OrderedObj[IdolPyCodegenTypeStruct]"
    ):
        self.fields = fields
        self.codegen_file = codegen_file
        super(IdolPyCodegenStruct, self).__init__(codegen_file.parent, codegen_file.path)

    @cached_property
    def declared_ident(self) -> Alt[Exported]:
        return Alt.lift(
            self.export(
                self.codegen_file.default_type_name,
                scripter.nameable_class_dec(
                    [self.import_ident(self.codegen_file.idol_py.idol_py_file.struct)],
                    [
                        line
                        for field_name, field in self.fields
                        for typing_expr in field.typing_expr
                        for line in (
                            scripter.comments(
                                get_tag_values(field.ts_decon.context.field_tags, "description")
                            )
                            + [
                                scripter.typing(
                                    get_safe_ident(field_name), self.apply_expr(typing_expr)
                                ),
                                "",
                            ]
                        )
                    ]
                    + [
                        scripter.assignment(
                            "__field_constructors__",
                            scripter.array(
                                scripter.tuple(
                                    scripter.literal(field_name),
                                    scripter.literal(get_safe_ident(field_name)),
                                    self.apply_expr(field_expr),
                                    scripter.invocation(
                                        "dict",
                                        optional=scripter.literal(
                                            includes_tag(
                                                field.ts_decon.context.field_tags, "optional"
                                            )
                                        ),
                                    ),
                                )
                                for field_name, field in self.fields
                                for field_expr in field.constructor_expr
                            ),
                        )
                    ],
                    doc_str=get_tag_values(self.codegen_file.t.tags, "description"),
                ),
            )
        )


class IdolPyCodegenEnum(GeneratorFileContext):
    codegen_file: IdolPyCodegenFile
    options: List[str]

    def __init__(self, codegen_file: IdolPyCodegenFile, options: List[str]):
        self.options = options
        self.codegen_file = codegen_file
        super(IdolPyCodegenEnum, self).__init__(codegen_file.parent, codegen_file.path)

    @cached_property
    def declared_ident(self) -> Alt[Exported]:
        return Alt.lift(
            self.export(
                self.codegen_file.default_type_name,
                scripter.nameable_class_dec(
                    [self.import_ident(self.codegen_file.idol_py.idol_py_file.enum)],
                    [
                        scripter.assignment(name.upper(), scripter.literal(name))
                        for name in self.options
                    ],
                    doc_str=get_tag_values(self.codegen_file.t.tags, "description"),
                ),
            )
        )


class IdolPyCodegenTypeStruct(GeneratorContext):
    ts_decon: TypeStructDeconstructor
    idol_py: IdolPy

    def __init__(self, parent: IdolPy, ts_decon: TypeStructDeconstructor):
        super(IdolPyCodegenTypeStruct, self).__init__(parent.state, parent.config)
        self.ts_decon = ts_decon
        self.idol_py = parent

    @cached_property
    def inner_scalar(self) -> "Alt[IdolPyCodegenScalar]":
        return Alt(
            IdolPyCodegenScalar(self.idol_py, scalar_decon)
            for scalar_decon in (
                self.ts_decon.get_scalar() ^ self.ts_decon.get_map() ^ self.ts_decon.get_repeated()
            )
        )

    @cached_property
    def typing_expr(self) -> Alt[Expression]:
        def repeated_expr(scalar_typing_expr: Expression) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.index_access(
                    state.import_ident(path, Exported(Path("typing"), "MutableSequence")),
                    scalar_typing_expr(state, path),
                )

            return inner

        def map_expr(scalar_typing_expr: Expression) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.index_access(
                    state.import_ident(path, Exported(Path("typing"), "MutableMapping")),
                    "str",
                    scalar_typing_expr(state, path),
                )

            return inner

        def optional_expr(inner_expr: Expression) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.index_access(
                    state.import_ident(path, Exported(Path("typing"), "Optional")),
                    inner_expr(state, path),
                )

            return inner

        container_typing: Alt[Expression] = Alt(
            repeated_expr(scalar_typing_expr)
            for scalar in self.inner_scalar
            for scalar_typing_expr in scalar.typing_expr
            if self.ts_decon.get_repeated()
        )

        container_typing ^= Alt(
            map_expr(scalar_typing_expr)
            for scalar in self.inner_scalar
            for scalar_typing_expr in scalar.typing_expr
            if self.ts_decon.get_map()
        )

        typing = container_typing + Alt(
            expr for scalar in self.inner_scalar for expr in scalar.typing_expr
        )

        if includes_tag(self.ts_decon.context.field_tags, "optional"):
            return Alt(optional_expr(typing_expr) for typing_expr in typing)

        return typing

    @cached_property
    def constructor_expr(self) -> Alt[Expression]:
        return self.container_constructor_expr ^ Alt(
            expr
            for inner_scalar in self.inner_scalar
            for expr in inner_scalar.constructor_expr
            if self.ts_decon.get_scalar()
        )

    @cached_property
    def container_constructor_expr(self) -> Alt[Expression]:
        container_con: Alt[Expression] = self.ts_decon.get_repeated() and Alt.lift(
            import_expr(self.idol_py.idol_py_file.list)
        )
        container_con ^= self.ts_decon.get_map() and Alt.lift(
            import_expr(self.idol_py.idol_py_file.map)
        )

        def container_expression(
            container_expr: Expression, scalar_expression: Expression
        ) -> Expression:
            def inner(state: GeneratorAcc, path: Path):
                return scripter.invocation(
                    scripter.prop_access(container_expr(state, path), "of"),
                    scalar_expression(state, path),
                    scripter.invocation(
                        "dict",
                        atleast_one=scripter.literal(
                            includes_tag(self.ts_decon.context.type_tags, "atleast_one")
                        ),
                    ),
                )

            return inner

        return Alt(
            container_expression(container_con_expr, scalar_con_expr)
            for container_con_expr in container_con
            for scalar in self.inner_scalar
            for scalar_con_expr in scalar.constructor_expr
        )


class IdolPyCodegenTypeStructDeclaration(IdolPyCodegenTypeStruct, AbstractGeneratorFileContext):
    path: Path
    codegen_file: IdolPyCodegenFile

    def __init__(self, parent: IdolPyCodegenFile, ts_decon: TypeStructDeconstructor):
        super(IdolPyCodegenTypeStructDeclaration, self).__init__(parent.idol_py, ts_decon)
        self.codegen_file = parent
        self.path = parent.path

    @cached_property
    def declarable_scalar(self) -> "Alt[IdolPyCodegenScalarDeclaration]":
        return Alt(
            IdolPyCodegenScalarDeclaration(self.codegen_file, scalar_decon)
            for scalar_decon in self.ts_decon.get_scalar()
        )

    @cached_property
    def declared_ident(self) -> Alt[Exported]:
        return Alt(
            self.export(
                self.codegen_file.default_type_name,
                scripter.commented(
                    get_tag_values(self.codegen_file.t_decon.t.tags, "description"),
                    scripter.declare_and_shadow(
                        self.apply_expr(typing_expr), self.apply_expr(constructor_expr)
                    ),
                ),
            )
            for typing_expr in self.typing_expr
            for constructor_expr in self.container_constructor_expr
        ) ^ Alt(
            declaration
            for scalar in self.declarable_scalar
            for declaration in scalar.declared_ident
        )


class IdolPyCodegenScalar(GeneratorContext):
    idol_py: IdolPy
    scalar_dec: ScalarDeconstructor

    def __init__(self, parent: IdolPy, scalar_decon: ScalarDeconstructor):
        super(IdolPyCodegenScalar, self).__init__(parent.state, parent.config)
        self.scalar_dec = scalar_decon
        self.idol_py = parent

    @cached_property
    def typing_expr(self) -> Alt[Expression]:
        return self.reference_import_expr ^ self.prim_typing_expr

    @cached_property
    def constructor_expr(self) -> Alt[Expression]:
        return self.reference_import_expr ^ self.prim_constructor_expr

    @cached_property
    def reference_import_expr(self) -> Alt[Expression]:
        alias_codegen_file = Alt(
            self.idol_py.codegen_file(ref) for ref in self.scalar_dec.get_alias()
        )

        alias_scaffold_file = Alt(
            self.idol_py.scaffold_file(ref)
            for ref in self.scalar_dec.get_alias()
            if ref.qualified_name in self.config.params.scaffold_types.obj
        )

        imported_alias_ident: Alt[Expression] = Alt(
            import_expr(codegen_type, "Codegen" + codegen_type.ident)
            for codegen_file in alias_codegen_file
            for codegen_type in codegen_file.declared_type_ident
            if not alias_scaffold_file
        )

        imported_alias_ident ^= Alt(
            import_expr(scaffold_type, "Scaffold" + scaffold_type.ident)
            for scaffold_file in alias_scaffold_file
            for scaffold_type in scaffold_file.declared_type_ident
        )

        return imported_alias_ident

    @cached_property
    def prim_constructor_expr(self) -> Alt[Expression]:
        def prim_con_expr(prim_expr: Expression) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.invocation(
                    scripter.prop_access(
                        state.import_ident(path, self.idol_py.idol_py_file.primitive), "of"
                    ),
                    prim_expr(state, path),
                )

            return inner

        def literal_con_expr(value: Any) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.invocation(
                    scripter.prop_access(
                        state.import_ident(path, self.idol_py.idol_py_file.literal), "of"
                    ),
                    scripter.literal(value),
                )

            return inner

        constructor_expr: Alt[Expression] = Alt(
            prim_con_expr(prim_expr)
            for prim_expr in self.prim_typing_expr
            if self.scalar_dec.get_primitive()
        )

        constructor_expr ^= Alt(
            literal_con_expr(value) for _, value in self.scalar_dec.get_literal()
        )

        return constructor_expr

    @cached_property
    def prim_typing_expr(self) -> Alt[Expression]:
        scalar_prim: Alt[PrimitiveType] = self.scalar_dec.get_primitive()
        scalar_prim ^= Alt(prim_type for prim_type, _ in self.scalar_dec.get_literal())

        def any_expr(state: GeneratorAcc, path: Path) -> str:
            return state.import_ident(path, Exported(Path("typing"), "Any"))

        def scalar_expr(prim_type: PrimitiveType) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return self.scalar_type_name_mappings[prim_type]

            return inner

        return Alt(
            scalar_expr(prim_type)
            for prim_type in scalar_prim
            if prim_type in self.scalar_type_name_mappings
        ) ^ Alt(any_expr for prim_type in scalar_prim if prim_type == PrimitiveType.ANY)

    scalar_type_name_mappings = {
        PrimitiveType.BOOL: "bool",
        PrimitiveType.INT: "int",
        PrimitiveType.STRING: "str",
        PrimitiveType.DOUBLE: "float",
    }


class IdolPyCodegenScalarDeclaration(IdolPyCodegenScalar, AbstractGeneratorFileContext):
    path: Path
    codegen_file: IdolPyCodegenFile

    def __init__(self, parent: IdolPyCodegenFile, scalar_decon: ScalarDeconstructor):
        super(IdolPyCodegenScalarDeclaration, self).__init__(parent.idol_py, scalar_decon)
        self.codegen_file = parent
        self.path = parent.path

    @cached_property
    def declared_prim_ident(self) -> Alt[Exported]:
        return Alt(
            self.export(
                self.codegen_file.default_type_name,
                scripter.commented(
                    get_tag_values(self.codegen_file.t_decon.t.tags, "description"),
                    scripter.declare_and_shadow(
                        self.apply_expr(prim_expr), self.apply_expr(prim_con_expr)
                    ),
                ),
            )
            for prim_expr in self.prim_typing_expr
            for prim_con_expr in self.prim_constructor_expr
        )

    @cached_property
    def declared_ident(self) -> Alt[Exported]:
        return self.declared_prim_ident ^ self.declared_alias_ident

    @cached_property
    def declared_alias_ident(self) -> Alt[Exported]:
        return Alt(
            self.export(
                self.codegen_file.default_type_name,
                scripter.commented(
                    get_tag_values(self.codegen_file.t_decon.t.tags, "description"),
                    scripter.assignable(self.apply_expr(ref_import)),
                ),
            )
            for ref_import in self.reference_import_expr
        )


class IdolPyScaffoldFile(GeneratorFileContext):
    t: Type
    idol_py: IdolPy

    def __init__(self, idol_py: IdolPy, t: Type, path: Path):
        self.t = t
        self.idol_py = idol_py
        super(IdolPyScaffoldFile, self).__init__(idol_py, path)

        self.reserve_ident(self.default_type_name)

    @cached_property
    def declared_type_ident(self) -> Alt[Exported]:
        type_decon = get_material_type_deconstructor(self.idol_py.config.params.all_types, self.t)

        codegen_type_ident: Alt[str] = Alt(
            self.import_ident(codegen_type, self.default_type_name + "Codegen")
            for codegen_type in self.idol_py.codegen_file(self.t.named).declared_type_ident
        )

        return Alt(
            self.export(self.default_type_name, scripter.assignable(codegen_ident))
            for codegen_ident in codegen_type_ident
            if type_decon.get_typestruct() or type_decon.get_enum()
        ) ^ Alt(
            self.export(
                self.default_type_name,
                scripter.nameable_class_dec(
                    [codegen_ident], [], doc_str=get_tag_values(self.t.tags, "description")
                ),
            )
            for codegen_ident in codegen_type_ident
            if type_decon.get_struct()
        )

    @property
    def default_type_name(self) -> str:
        return self.t.named.type_name


class IdolPyFile(ExternFileContext):
    idol_py: IdolPy

    EXTERN_FILE = os.path.join(os.path.dirname(__file__), "__idol__.py")

    def __init__(self, idol_py: IdolPy, path: Path):
        self.idol_py = idol_py
        super(IdolPyFile, self).__init__(idol_py, path)

    @cached_property
    def list(self) -> Exported:
        return self.export_extern("List")

    @cached_property
    def map(self) -> Exported:
        return self.export_extern("Map")

    @cached_property
    def primitive(self) -> Exported:
        return self.export_extern("Primitive")

    @cached_property
    def literal(self) -> Exported:
        return self.export_extern("Literal")

    @cached_property
    def enum(self) -> Exported:
        return self.export_extern("Enum")

    @cached_property
    def struct(self) -> Exported:
        return self.export_extern("Struct")


def main():
    params: GeneratorParams = start(
        CliConfig(
            args={
                "target": "idol module names whose contents will have extensible types scaffolded.",
                "output": "a directory to generate the scaffolds and codegen into.",
            }
        )
    )

    config = GeneratorConfig(params)
    config.with_path_mappings(
        dict(
            codegen=config.in_codegen_dir(config.one_file_per_type),
            scaffold=config.one_file_per_type,
        )
    )

    idol_py = IdolPy(config)
    move_to = build(config, idol_py.render())
    move_to(params.output_dir)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import os.path

from cached_property import cached_property
from typing import Callable, List, Dict, Union, Any, Optional

from idol import scripter
from idol.cli import start, CliConfig
from idol.functional import OrderedObj, Alt
from idol.generator import (
    GeneratorParams,
    GeneratorConfig,
    build,
    GeneratorAcc,
    TypeDeconstructor,
    Path,
    get_material_type_deconstructor,
    TypeStructDeconstructor,
    ScalarDeconstructor,
    GeneratorContext,
    GeneratorFileContext,
    Exported,
    Expression,
    import_expr,
    get_safe_ident,
    get_tag_values,
    includes_tag,
    TypeStructContext,
    ExternFileContext,
    ImportPath,
    AbstractGeneratorFileContext,
)
from idol.py.schema.primitive_type import PrimitiveType
from idol.py.schema.reference import Reference
from idol.py.schema.type import Type


class FieldExpressionComposer:
    constructor_ident: Expression
    args: List[Union[Expression, str]]
    kwds: Dict[str, Union[str, str]]

    def __init__(
        self,
        constructor_ident: Expression,
        args: List[Union[str, Expression]] = None,
        kwds: Dict[str, Union[str, Expression]] = None,
    ):
        self.constructor_ident = constructor_ident
        self.args = args or []
        self.kwds = kwds or {}

    def resolve_arg_or_kwd(
        self, expr_or_str: Union[Expression, str], state: GeneratorAcc, path: Path
    ):
        if isinstance(expr_or_str, str):
            return expr_or_str

        return expr_or_str(state, path)

    def resolve_args_and_kwds(self, state: GeneratorAcc, path: Path):
        args = [self.resolve_arg_or_kwd(expr_or_str, state, path) for expr_or_str in self.args]
        kwds = {
            k: self.resolve_arg_or_kwd(expr_or_str, state, path)
            for k, expr_or_str in self.kwds.items()
        }
        return args, kwds

    def field_instance_expr(self) -> Expression:
        def expr(state: GeneratorAcc, path: Path) -> str:
            args, kwds = self.resolve_args_and_kwds(state, path)
            return scripter.invocation(self.constructor_ident(state, path), *args, **kwds)

        return expr

    def curried_field_declaration_expr(self, name: str, idol_mar: "IdolMarshmallow") -> Expression:
        def expr(state: GeneratorAcc, path: Path) -> str:
            args, kwds = self.resolve_args_and_kwds(state, path)
            return scripter.invocation(
                state.import_ident(path, idol_mar.idol_mar_file.wrap_field),
                self.constructor_ident(state, path),
                scripter.literal(name),
                *list(scripter.thunk(arg) for arg in args),
                **kwds,
            )

        return expr

    def with_more_kwds(
        self, more_kwds: Dict[str, Union[Expression, str]]
    ) -> "FieldExpressionComposer":
        return FieldExpressionComposer(
            self.constructor_ident, self.args, dict(**self.kwds, **more_kwds)
        )


class IdolMarCodegenStruct(GeneratorFileContext):
    codegen_file: "IdolMarCodegenFile"
    fields: OrderedObj["IdolMarCodegenTypeStruct"]

    def __init__(
        self, codegen_file: "IdolMarCodegenFile", fields: "OrderedObj[IdolMarCodegenTypeStruct]"
    ):
        self.fields = fields
        self.codegen_file = codegen_file
        super(IdolMarCodegenStruct, self).__init__(codegen_file.parent, codegen_file.path)

    @cached_property
    def declared_schema(self) -> Alt[Exported]:
        return Alt.lift(
            self.export(
                self.codegen_file.default_schema_name,
                scripter.nameable_class_dec(
                    [self.import_ident(Exported(Path("marshmallow"), "Schema"))],
                    [
                        line
                        for field_name, field in self.fields
                        for composer in field.field_composer
                        for line in (
                            scripter.comments(
                                get_tag_values(field.ts_decon.context.field_tags, "description")
                            )
                            + [
                                scripter.assignment(
                                    get_safe_ident(field_name),
                                    self.apply_expr(
                                        composer.with_more_kwds(
                                            dict(
                                                dump_to=scripter.literal(field_name),
                                                load_from=scripter.literal(field_name),
                                                allow_none=scripter.literal(
                                                    includes_tag(
                                                        field.ts_decon.context.field_tags,
                                                        "optional",
                                                    )
                                                ),
                                            )
                                        ).field_instance_expr()
                                    ),
                                )
                            ]
                        )
                    ],
                    doc_str=get_tag_values(self.codegen_file.t.tags, "description"),
                ),
            )
        )

    @cached_property
    def declared_field(self) -> Alt[Exported]:
        for declared_schema in self.declared_schema:
            schema_expr: Expression = import_expr(declared_schema)

            if self.codegen_file.t.named.qualified_name in self.config.params.scaffold_types.obj:
                for scaffold_declared_schema in self.codegen_file.idol_mar.scaffold_file(
                    self.codegen_file.t.named
                ).declared_schema:

                    def lazy_import(state: GeneratorAcc, path: Path) -> str:
                        import_module = state.import_ident(
                            path, Exported(Path("importlib"), "import_module")
                        )

                        return scripter.prop_access(
                            scripter.invocation(
                                import_module,
                                scripter.literal(
                                    ImportPath.as_python_module_path(
                                        path.import_path_to(scaffold_declared_schema.path).rel_path
                                    )
                                ),
                                "__package__",
                            ),
                            scaffold_declared_schema.ident,
                        )

                    schema_expr = lazy_import

            return Alt.lift(
                self.export(
                    self.codegen_file.default_field_name,
                    scripter.assignable(
                        self.apply_expr(
                            FieldExpressionComposer(
                                import_expr(
                                    self.codegen_file.idol_mar.marshmallow_field_exported("Nested")
                                ),
                                [self.apply_expr(schema_expr)],
                            ).curried_field_declaration_expr(
                                self.codegen_file.default_field_name, self.codegen_file.idol_mar
                            )
                        )
                    ),
                )
            )

        return Alt.empty()


class IdolMarCodegenEnum(GeneratorFileContext):
    codegen_file: "IdolMarCodegenFile"
    options: List[str]

    def __init__(self, codegen_file: "IdolMarCodegenFile", options: List[str]):
        self.options = options
        self.codegen_file = codegen_file
        super(IdolMarCodegenEnum, self).__init__(codegen_file.parent, codegen_file.path)

        self.reserve_ident(self.default_enum_name)

    @cached_property
    def declared_enum(self) -> Alt[Exported]:
        return Alt.lift(
            self.export(
                self.default_enum_name,
                scripter.nameable_class_dec(
                    [self.import_ident(Exported(Path("enum"), "Enum"))],
                    [
                        scripter.assignment(name.upper(), scripter.literal(name))
                        for name in self.options
                    ],
                    doc_str=get_tag_values(self.codegen_file.t.tags, "description"),
                ),
            )
        )

    @cached_property
    def declared_field(self) -> Alt[Exported]:
        return Alt(
            self.export(
                self.codegen_file.default_field_name,
                scripter.assignable(
                    self.apply_expr(
                        FieldExpressionComposer(
                            import_expr(Exported(Path("marshmallow_enum"), "EnumField")),
                            [self.import_ident(declared_enum)],
                            dict(by_value=scripter.literal(True)),
                        ).curried_field_declaration_expr(
                            self.codegen_file.default_field_name, self.codegen_file.idol_mar
                        )
                    )
                ),
            )
            for declared_enum in self.declared_enum
        )

    @cached_property
    def declared_schema(self) -> Alt[Exported]:
        return Alt.lift(
            self.export(
                self.codegen_file.default_schema_name, scripter.assignable(scripter.literal(None))
            )
        )

    @property
    def default_enum_name(self):
        return self.codegen_file.t.named.as_qualified_ident + "Enum"


class IdolMarCodegenScalar(GeneratorContext):
    idol_mar: "IdolMarshmallow"
    scalar_dec: ScalarDeconstructor

    FieldExpressionComposer = FieldExpressionComposer

    def __init__(self, parent: "IdolMarshmallow", scalar_decon: ScalarDeconstructor):
        super(IdolMarCodegenScalar, self).__init__(parent.state, parent.config)
        self.scalar_dec = scalar_decon
        self.idol_mar = parent

    @cached_property
    def field_expr_composer(self) -> Alt[FieldExpressionComposer]:
        return (
            self.reference_import_field_expr_composer
            ^ self.prim_field_expr
            ^ self.literal_field_expr
        )

    @cached_property
    def reference_import_field_expr_composer(self) -> Alt[FieldExpressionComposer]:
        return Alt(
            self.FieldExpressionComposer(import_expr(declared_field))
            for ref in self.scalar_dec.get_alias()
            for declared_field in self.idol_mar.codegen_file(ref).declared_field
        )

    @cached_property
    def literal_field_expr(self) -> Alt[FieldExpressionComposer]:
        return Alt(
            self.FieldExpressionComposer(
                import_expr(self.idol_mar.marshmallow_field_exported("Constant")),
                [scripter.literal(val)],
            )
            for _, val in self.scalar_dec.get_literal()
        )

    @cached_property
    def prim_field_expr(self) -> Alt[FieldExpressionComposer]:
        scalar_prim: Alt[PrimitiveType] = self.scalar_dec.get_primitive()

        return Alt(
            FieldExpressionComposer(import_expr(self.scalar_marshmallow_field_mappings[prim_type]))
            for prim_type in scalar_prim
            if prim_type in self.scalar_marshmallow_field_mappings
        ) ^ Alt(
            FieldExpressionComposer(import_expr(self.idol_mar.idol_mar_file.any))
            for prim_type in scalar_prim
            if prim_type == PrimitiveType.ANY
        )

    @cached_property
    def scalar_marshmallow_field_mappings(self) -> Dict[PrimitiveType, Exported]:
        return {
            PrimitiveType.BOOL: self.idol_mar.marshmallow_field_exported("Boolean"),
            PrimitiveType.INT: self.idol_mar.marshmallow_field_exported("Int"),
            PrimitiveType.STRING: self.idol_mar.marshmallow_field_exported("String"),
            PrimitiveType.DOUBLE: self.idol_mar.marshmallow_field_exported("Float"),
        }

    @cached_property
    def scalar_defaults_mapping(self) -> Dict[PrimitiveType, str]:
        return {
            PrimitiveType.BOOL: scripter.literal(False),
            PrimitiveType.INT: scripter.literal(0),
            PrimitiveType.STRING: scripter.literal(""),
            PrimitiveType.DOUBLE: scripter.literal(0.0),
            PrimitiveType.ANY: scripter.literal(0),
        }


class IdolMarCodegenTypeStruct(GeneratorContext):
    ts_decon: TypeStructDeconstructor
    idol_mar: "IdolMarshmallow"

    IdolMarCodegenScalar = IdolMarCodegenScalar

    def __init__(self, parent: "IdolMarshmallow", ts_decon: TypeStructDeconstructor):
        super(IdolMarCodegenTypeStruct, self).__init__(parent.state, parent.config)
        self.ts_decon = ts_decon
        self.idol_mar = parent

    @cached_property
    def inner_scalar(self) -> "Alt[IdolMarCodegenScalar]":
        return Alt(
            self.IdolMarCodegenScalar(self.idol_mar, scalar_decon)
            for scalar_decon in (
                self.ts_decon.get_scalar() ^ self.ts_decon.get_map() ^ self.ts_decon.get_repeated()
            )
        )

    @cached_property
    def field_composer(self) -> Alt[FieldExpressionComposer]:
        return self.scalar_field_composer ^ self.list_field_composer ^ self.map_field_composer

    @cached_property
    def scalar_field_composer(self) -> Alt["FieldExpressionComposer"]:
        return Alt(
            composer
            for scalar in self.inner_scalar
            for composer in scalar.field_expr_composer
            if self.ts_decon.get_scalar()
        )

    @cached_property
    def list_field_composer(self) -> Alt["FieldExpressionComposer"]:
        return Alt(
            FieldExpressionComposer(
                import_expr(self.idol_mar.marshmallow_field_exported("List")),
                [field_composer.field_instance_expr()],
            )
            for scalar in self.inner_scalar
            for field_composer in scalar.field_expr_composer
            if self.ts_decon.get_repeated()
        )

    @cached_property
    def map_field_composer(self) -> Alt["FieldExpressionComposer"]:
        if self.ts_decon.get_map():
            return Alt.lift(
                FieldExpressionComposer(
                    import_expr(self.idol_mar.marshmallow_field_exported("Dict"))
                )
            )
        return Alt.empty()


class IdolMarCodegenTypeStructDeclaration(IdolMarCodegenTypeStruct, AbstractGeneratorFileContext):
    path: Path
    codegen_file: "IdolMarCodegenFile"

    def __init__(self, parent: "IdolMarCodegenFile", ts_decon: TypeStructDeconstructor):
        super(IdolMarCodegenTypeStructDeclaration, self).__init__(parent.idol_mar, ts_decon)
        self.codegen_file = parent
        self.path = parent.path

    @cached_property
    def declared_field(self) -> Alt[Exported]:
        return Alt(
            self.export(
                self.codegen_file.default_field_name,
                scripter.assignable(
                    self.apply_expr(
                        composer.curried_field_declaration_expr(
                            self.codegen_file.default_field_name, self.idol_mar
                        )
                    )
                ),
            )
            for composer in self.field_composer
        )

    @cached_property
    def declared_schema(self) -> Alt[Exported]:
        for scalar in self.ts_decon.get_scalar():
            for ref in scalar.get_alias():
                schema: Alt[Exported] = (
                    self.codegen_file.idol_mar.scaffold_file(ref).declared_schema
                    if ref.qualified_name in self.config.params.scaffold_types.obj
                    else self.codegen_file.idol_mar.codegen_file(ref).declared_schema
                )

                for exported_schema in schema:
                    return Alt.lift(
                        self.export(
                            self.codegen_file.default_schema_name,
                            scripter.assignable(self.import_ident(exported_schema)),
                        )
                    )

        return Alt.lift(
            self.export(
                self.codegen_file.default_schema_name, scripter.assignable(scripter.literal(None))
            )
        )


class IdolMarScaffoldFile(GeneratorFileContext):
    t: Type
    idol_mar: "IdolMarshmallow"

    def __init__(self, idol_mar: "IdolMarshmallow", t: Type, path: Path):
        self.t = t
        self.idol_mar = idol_mar
        super(IdolMarScaffoldFile, self).__init__(idol_mar, path)

        self.reserve_ident(self.default_schema_name)

    @cached_property
    def declared_schema(self) -> Alt[Exported]:
        type_decon = get_material_type_deconstructor(self.idol_mar.config.params.all_types, self.t)

        codegen_schema_ident: Alt[str] = Alt(
            self.import_ident(codegen_schema, self.default_schema_name + "Codegen")
            for codegen_schema in self.idol_mar.codegen_file(self.t.named).declared_schema
            if type_decon.get_struct()
        )

        return Alt(
            self.export(
                self.default_schema_name,
                scripter.nameable_class_dec(
                    [codegen_ident], [], doc_str=get_tag_values(self.t.tags, "description")
                ),
            )
            for codegen_ident in codegen_schema_ident
            if type_decon.get_struct()
        )

    @property
    def default_schema_name(self) -> str:
        return self.t.named.type_name + "Schema"


class IdolMarFile(ExternFileContext):
    idol_mar: "IdolMarshmallow"

    EXTERN_FILE = os.path.join(os.path.dirname(__file__), "__idol_mar__.py")

    def __init__(self, idol_mar: "IdolMarshmallow", path: Path):
        self.idol_mar = idol_mar
        super(IdolMarFile, self).__init__(idol_mar, path)

    @cached_property
    def any(self) -> Exported:
        return self.export_extern("Any")

    @cached_property
    def wrap_field(self) -> Exported:
        return self.export_extern("wrap_field")


class IdolMarCodegenFile(GeneratorFileContext):
    t: Type
    idol_mar: "IdolMarshmallow"
    t_decon: TypeDeconstructor

    IdolMarCodegenTypeStructDeclaration = IdolMarCodegenTypeStructDeclaration
    IdolMarCodegenEnum = IdolMarCodegenEnum
    IdolMarCodegenStruct = IdolMarCodegenStruct
    IdolMarCodegenTypeStruct = IdolMarCodegenTypeStruct

    def __init__(self, idol_mar: "IdolMarshmallow", t: Type, path: Path):
        self.t = t
        self.idol_mar = idol_mar
        self.t_decon = TypeDeconstructor(t)
        super(IdolMarCodegenFile, self).__init__(idol_mar, path)

        self.reserve_ident(self.default_schema_name)
        self.reserve_ident(self.default_field_name)

    @cached_property
    def declared_field(self) -> Alt[Exported]:
        declared_field: Alt[Exported] = Alt(
            v for typestruct in self.typestruct for v in typestruct.declared_field
        )

        declared_field ^= Alt(v for enum in self.enum for v in enum.declared_field)
        declared_field ^= Alt(v for struct in self.struct for v in struct.declared_field)

        return declared_field

    @cached_property
    def declared_schema(self) -> Alt[Exported]:
        declared_schema: Alt[Exported] = Alt(
            v for typestruct in self.typestruct for v in typestruct.declared_schema
        )

        declared_schema ^= Alt(v for enum in self.enum for v in enum.declared_schema)
        declared_schema ^= Alt(v for struct in self.struct for v in struct.declared_schema)

        return declared_schema

    @cached_property
    def declared_default_factory(self) -> Alt[Exported]:
        return Alt.empty()

    @property
    def default_default_factory_name(self) -> str:
        return f"{self.t.named.as_qualified_ident}Default"

    @property
    def default_field_name(self) -> str:
        return f"{self.t.named.as_qualified_ident}Field"

    @property
    def default_schema_name(self) -> str:
        return f"{self.t.named.as_qualified_ident}Schema"

    @cached_property
    def typestruct(self) -> "Alt[IdolMarCodegenTypeStructDeclaration]":
        return Alt(
            self.IdolMarCodegenTypeStructDeclaration(self, ts_decon)
            for ts_decon in self.t_decon.get_typestruct()
        )

    @cached_property
    def enum(self) -> "Alt[IdolMarCodegenEnum]":
        return Alt(self.IdolMarCodegenEnum(self, options) for options in self.t_decon.get_enum())

    @cached_property
    def struct(self) -> "Alt[IdolMarCodegenStruct]":
        return Alt(
            self.IdolMarCodegenStruct(
                self,
                OrderedObj.from_iterable(
                    OrderedObj({k: self.IdolMarCodegenTypeStruct(self.idol_mar, ts_decon)})
                    for k, ts_decon in fields
                ),
            )
            for fields in self.t_decon.get_struct()
        )


class IdolMarshmallow(GeneratorContext):
    scaffolds: Dict[str, "IdolMarScaffoldFile"]
    codegens: Dict[str, "IdolMarCodegenFile"]

    IdolMarScaffoldFile = IdolMarScaffoldFile
    IdolMarCodegenFile = IdolMarCodegenFile
    IdolMarFile = IdolMarFile

    def __init__(self, config: GeneratorConfig):
        super(IdolMarshmallow, self).__init__(GeneratorAcc(), config)
        self.scaffolds = {}
        self.codegens = {}

    def scaffold_file(self, ref: Reference) -> "IdolMarScaffoldFile":
        path = self.state.reserve_path(**self.config.paths_of(scaffold=ref))
        t = self.config.params.all_types.obj[ref.qualified_name]

        if t.named.qualified_name not in self.scaffolds:
            self.scaffolds[t.named.qualified_name] = self.IdolMarScaffoldFile(self, t, path)

        return self.scaffolds[t.named.qualified_name]

    def codegen_file(self, ref: Reference) -> "IdolMarCodegenFile":
        path = self.state.reserve_path(**self.config.paths_of(codegen=ref))
        t = self.config.params.all_types.obj[ref.qualified_name]

        if t.named.qualified_name not in self.codegens:
            self.codegens[t.named.qualified_name] = self.IdolMarCodegenFile(self, t, path)

        return self.codegens[ref.qualified_name]

    @cached_property
    def idol_mar_file(self) -> "IdolMarFile":
        path = self.state.reserve_path(runtime=self.config.codegen_root + "/__idol__.py")
        return self.IdolMarFile(self, path)

    def render(self) -> OrderedObj[str]:
        for i, t in enumerate(self.config.params.scaffold_types.values()):
            for export in self.scaffold_file(t.named).declared_schema:
                print(
                    f"Rendered {export.ident} to {export.path.path} ({i} / {len(self.config.params.scaffold_types)})"
                )
                break
            else:
                print(
                    f"Skipped {t.named.as_qualified_ident} ({i} / {len(self.config.params.scaffold_types)})"
                )

        return self.state.render(
            dict(
                codegen=[
                    "DO NOT EDIT",
                    "This file was generated by idol_mar, any changes will be lost when idol_mar is rerun again",
                ],
                scaffold=[
                    "This file was scaffold by idol_mar, but it will not be overwritten, so feel free to edit.",
                    "This file will be regenerated if you delete it.",
                ],
            )
        )

    @classmethod
    def marshmallow_field_exported(cls, ident: str) -> Exported:
        return Exported(Path("marshmallow.fields"), ident)


def main():
    params: GeneratorParams = start(
        CliConfig(
            args={
                "target": "idol module names whose contents will have extensible types scaffolded.",
                "output": "a directory to generate the scaffolds and codegen into.",
            }
        )
    )

    config = GeneratorConfig(params)
    config.with_path_mappings(
        dict(
            codegen=config.in_codegen_dir(config.one_file_per_type),
            scaffold=config.one_file_per_type,
        )
    )

    idol_mar = IdolMarshmallow(config)
    move_to = build(config, idol_mar.render())
    move_to(params.output_dir)


if __name__ == "__main__":
    main()



# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Struct, Primitive


class SchemaLiteral(Struct):
    bool: bool

    double: float

    int: int

    string: str

    __field_constructors__ = [
        ("bool", "bool", Primitive.of(bool), dict(optional=False),),
        ("double", "double", Primitive.of(float), dict(optional=False),),
        ("int", "int", Primitive.of(int), dict(optional=False),),
        ("string", "string", Primitive.of(str), dict(optional=False),),
    ]



# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Struct, Primitive, List
from typing import MutableSequence
from ...schema.type_struct import TypeStruct as ScaffoldTypeStruct


class SchemaField(Struct):
    field_name: str

    tags: MutableSequence[str]

    type_struct: ScaffoldTypeStruct

    __field_constructors__ = [
        ("field_name", "field_name", Primitive.of(str), dict(optional=False),),
        (
            "tags",
            "tags",
            List.of(Primitive.of(str), dict(atleast_one=False)),
            dict(optional=False),
        ),
        ("type_struct", "type_struct", ScaffoldTypeStruct, dict(optional=False),),
    ]



from idol.functional import OrderedObj
from ..codegen.schema.module import SchemaModule
from .type import Type


class Module(SchemaModule):
    def types_as_ordered_obj(self) -> OrderedObj[Type]:
        return OrderedObj.from_iterable(
            OrderedObj({t.named.qualified_name: t})
            for n in self.types_dependency_ordering
            for t in [self.types_by_name[n]]
        )



# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Struct
from typing import Optional
from ...schema.literal import Literal as ScaffoldLiteral
from ...schema.primitive_type import PrimitiveType as ScaffoldPrimitiveType
from ...schema.reference import Reference as ScaffoldReference
from ...schema.struct_kind import StructKind as ScaffoldStructKind


class SchemaTypeStruct(Struct):
    literal: Optional[ScaffoldLiteral]

    primitive_type: ScaffoldPrimitiveType

    reference: ScaffoldReference

    struct_kind: ScaffoldStructKind

    __field_constructors__ = [
        ("literal", "literal", ScaffoldLiteral, dict(optional=True),),
        (
            "primitive_type",
            "primitive_type",
            ScaffoldPrimitiveType,
            dict(optional=False),
        ),
        ("reference", "reference", ScaffoldReference, dict(optional=False),),
        ("struct_kind", "struct_kind", ScaffoldStructKind, dict(optional=False),),
    ]



import tempfile
import os.path
import os
import shutil


class BuildEnv:
    codegen_root: str
    build_dir: str

    def __init__(self, name: str, codegen_root: str):
        self.build_dir = tempfile.mkdtemp(prefix=name)
        self.codegen_root = codegen_root

    def abs_path(self, rel_path):
        return os.path.abspath(os.path.join(self.build_dir, rel_path))

    def write_build_file(self, rel_path: str, contents: str):
        abs_path = self.abs_path(rel_path)
        os.makedirs(os.path.dirname(abs_path), exist_ok=True)
        with open(abs_path, "w") as file:
            file.write(contents)

    def finalize(self, output_dir: str, replace: bool = False):
        # Create module files.
        for (dirpath, dirnames, _) in os.walk(self.build_dir):
            for name in dirnames:
                expected_init = os.path.join(dirpath, name, "__init__.py")
                expected_init = self.abs_path(expected_init)
                if not os.path.exists(expected_init):
                    with open(expected_init, "w") as file:
                        file.write("\n")

        existing_codegen = os.path.join(output_dir, self.codegen_root)
        if os.path.exists(existing_codegen):
            shutil.rmtree(existing_codegen)

        recursive_copy(self.build_dir, output_dir, replace)


def recursive_copy(src: str, dest: str, replace: bool):
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest, exist_ok=True)
        files = os.listdir(src)
        for f in files:
            recursive_copy(os.path.join(src, f), os.path.join(dest, f), replace)
    else:
        if not replace and os.path.exists(dest):
            print("Skipping", dest, "...")
        else:
            shutil.copyfile(src, dest)



from cached_property import cached_property

import idol.scripter as scripter
from typing import Dict, List, Union, TypeVar, Callable, Tuple, Any, Optional, Iterable

from idol.py.schema.primitive_type import PrimitiveType
from idol.py.schema.reference import Reference
from idol.py.schema.struct_kind import StructKind
from idol.py.schema.type import Type
from .build_env import BuildEnv
from .functional import OrderedObj, Alt, StringSet, naive_object_concat
from idol.py.schema.module import Module
from idol.py.schema.type_struct import TypeStruct
import os.path

A = TypeVar("A")


class Path:
    path: str

    def __init__(self, path: str):
        self.path = path

    def __eq__(self, other):
        return self.path == other.path

    @property
    def is_module(self):
        return not self.path.endswith(".py")

    def import_path_to(self, to_path: "Path") -> "ImportPath":
        # Special case for relative path to self => is_local
        if self == to_path:
            return ImportPath(to_path, "")

        if to_path.is_module:
            return ImportPath.module(to_path.path)

        if self.is_module:
            raise ValueError("Absolute modules cannot express relative import paths!")

        from_parts = self.path.split("/")
        to_parts = to_path.path.split("/")
        parts = []
        i = len(from_parts)

        while i > 0 and from_parts[:i] != to_parts[:i]:
            parts.append("..")
            i -= 1

        while i < len(to_parts):
            parts.append(to_parts[i])
            i += 1

        return ImportPath(to_path, "/".join(parts))

    def __str__(self):
        return self.path


class ImportPath:
    path: Path
    rel_path: str

    def __init__(self, path: Path, rel_path: str):
        self.path = path
        self.rel_path = rel_path

    def __str__(self):
        return str(self.path)

    @classmethod
    def module(cls, module: str) -> "ImportPath":
        return cls(Path(module), module)

    @property
    def is_module(self):
        return self.path.is_module

    @staticmethod
    def as_python_module_path(rel_path: str) -> str:
        if rel_path == "":
            raise ValueError("is_local path cannot be a python module path!")

        if rel_path.endswith(".py"):
            if rel_path[0] != ".":
                rel_path = "./" + rel_path
            return rel_path[:-3].replace("../", ".").replace("/", ".")

        return rel_path


class Exported:
    path: Path
    ident: str
    source_state: Optional["GeneratorAcc"]

    def __init__(self, path: Path, ident: str, source_state: Optional["GeneratorAcc"] = None):
        self.ident = ident
        self.path = path

        if not path.is_module and not source_state:
            raise ValueError(f"source_state parameter required for non module {path}")

        self.source_state = source_state


class GeneratorParams:
    def __init__(
        self,
        all_modules: OrderedObj[Module],
        all_types: OrderedObj[Type],
        scaffold_types: OrderedObj[Type],
        output_dir: str,
        options: Dict[str, Union[List[str], bool]],
    ):
        self.all_modules = all_modules
        self.all_types = all_types
        self.scaffold_types = scaffold_types
        self.output_dir = output_dir
        self.options = options


def includes_tag(tags: Optional[Iterable[str]], tag: str) -> bool:
    return bool(tags) and tag in list(tags)


def get_tag_value(tags: Optional[Iterable[str]], d: str, tag: str) -> str:
    if not tags:
        return d

    for t in tags:
        if t.startswith(tag + ":"):
            return t[len(tag) + 1 :]

    return d


def get_tag_values(tags: Optional[Iterable[str]], tag: str) -> List[str]:
    if not tags:
        return []

    result: List[str] = []

    for t in tags:
        if t.startswith(tag + ":"):
            result.append(t[len(tag) + 1 :])

    return result


class TypeStructContext:
    field_tags: List[str]
    type_tags: List[str]
    is_type_bound: bool
    type_display_name: str
    field_name: Optional[str]

    @property
    def is_declarable(self):
        return self.is_type_bound

    def __init__(
        self,
        type_display_name: str,
        field_name: Optional[str] = None,
        field_tags: Optional[List[str]] = None,
        type_tags: Optional[List[str]] = None,
    ):
        self.type_display_name = type_display_name
        self.field_name = field_name
        self.field_tags = field_tags or []
        self.type_tags = type_tags or []
        self.is_type_bound = type_tags is not None and field_tags is None


class ScalarContext:
    is_contained: bool
    typestruct_context: TypeStructContext

    def __init__(self, typestruct_context: TypeStructContext, is_contained: bool):
        self.typestruct_context = typestruct_context
        self.is_contained = is_contained

    @property
    def is_type_bound(self):
        return self.typestruct_context.is_type_bound

    @property
    def is_declarable(self):
        return not self.is_contained and self.is_type_bound


class ScalarDeconstructor:
    type_struct: TypeStruct
    context: ScalarContext

    def __init__(self, type_struct: TypeStruct, context: ScalarContext):
        self.type_struct = type_struct
        self.context = context

    def get_primitive(self) -> Alt[PrimitiveType]:
        if self.type_struct.is_alias or self.type_struct.is_literal:
            return Alt.empty()

        return Alt.lift(self.type_struct.primitive_type)

    def get_literal(self) -> Alt[Tuple[PrimitiveType, Any]]:
        if self.type_struct.is_alias or not self.type_struct.is_literal:
            return Alt.empty()

        return Alt.lift((self.type_struct.primitive_type, self.type_struct.literal_value))

    def get_alias(self) -> Alt[Reference]:
        if not self.type_struct.is_alias:
            return Alt.empty()

        return Alt.lift(self.type_struct.reference)


class TypeStructDeconstructor:
    type_struct: TypeStruct
    context: TypeStructContext

    def __init__(self, type_struct: TypeStruct, context: TypeStructContext):
        self.type_struct = type_struct
        self.context = context

    def get_scalar(self) -> Alt[ScalarDeconstructor]:
        if self.type_struct.struct_kind != StructKind.SCALAR:
            return Alt.empty()

        return Alt.lift(ScalarDeconstructor(self.type_struct, ScalarContext(self.context, False)))

    def get_repeated(self) -> Alt[ScalarDeconstructor]:
        if self.type_struct.struct_kind != StructKind.REPEATED:
            return Alt.empty()

        return Alt.lift(ScalarDeconstructor(self.type_struct, ScalarContext(self.context, False)))

    def get_map(self) -> Alt[ScalarDeconstructor]:
        if self.type_struct.struct_kind != StructKind.MAP:
            return Alt.empty()

        return Alt.lift(ScalarDeconstructor(self.type_struct, ScalarContext(self.context, False)))


class TypeDeconstructor:
    t: Type

    def __init__(self, t: Type):
        self.t = t

    def get_typestruct(self) -> Alt[TypeStructDeconstructor]:
        if not self.t.is_a:
            return Alt.empty()

        return Alt.lift(
            TypeStructDeconstructor(
                self.t.is_a,
                TypeStructContext(
                    type_display_name=self.t.named.qualified_name, type_tags=list(self.t.tags)
                ),
            )
        )

    def get_enum(self) -> Alt[List[str]]:
        if self.t.is_a or not self.t.is_enum:
            return Alt.empty()

        return Alt.lift(self.t.options)

    def get_struct(self) -> Alt[OrderedObj[TypeStructDeconstructor]]:
        if self.t.is_a or self.t.is_enum:
            return Alt.empty()

        return Alt.lift(
            OrderedObj(
                {
                    k: TypeStructDeconstructor(
                        v.type_struct,
                        TypeStructContext(
                            type_display_name=v.type_struct.type_display_name,
                            field_name=v.field_name,
                            field_tags=list(v.tags),
                        ),
                    )
                    for k, v in self.t.fields.items()
                }
            )
        )


def import_expr(exported: Exported, as_ident: str = None) -> "Expression":
    def inner(state: GeneratorAcc, path: Path) -> str:
        return state.import_ident(path, exported, as_ident)

    return inner


def as_expression(expr: str) -> "Expression":
    def inner(state: GeneratorAcc, path: Path) -> str:
        return expr

    return inner


def get_material_type_deconstructor(all_types: OrderedObj[Type], t: Type) -> TypeDeconstructor:
    def search_type(type_decon: TypeDeconstructor) -> TypeDeconstructor:
        return Alt(
            search_type(TypeDeconstructor(all_types.obj[alias.qualified_name].with_tags(t.tags)))
            for type_struct in type_decon.get_typestruct()
            for scalar in type_struct.get_scalar()
            for alias in scalar.get_alias()
        ).get_or(type_decon)

    return search_type(TypeDeconstructor(t))


class GeneratorConfig:
    codegen_root: str
    name: str
    path_mappings: Dict[str, Callable[[Reference], str]]
    params: GeneratorParams

    def __init__(self, params: GeneratorParams):
        self.params = params
        self.codegen_root = "codegen"
        self.name = "idol_py"
        self.path_mappings = {}

    def paths_of(self, **k: Dict[str, Reference]) -> Dict[str, str]:
        return {group: self.path_mappings[group](ref) for group, ref in k.items()}

    @staticmethod
    def one_file_per_type(r: Reference) -> str:
        return r.as_qn_path

    @staticmethod
    def one_file_per_module(r: Reference) -> str:
        return r.as_module_path

    @staticmethod
    def flat_namespace(r: Reference) -> str:
        return r.as_type_path

    def in_codegen_dir(self, m: Callable[[Reference], str]) -> Callable[[Reference], str]:
        def in_codegen_dir(r: Reference):
            return f"{self.codegen_root}/{m(r)}"

        return in_codegen_dir

    def with_path_mappings(self, path_mappings: Dict[str, Callable[[Reference], str]]):
        self.path_mappings = path_mappings


class IdentifiersAcc:
    # path -> IdentityName -> sources mset
    idents: OrderedObj[OrderedObj[StringSet]]

    def __init__(self):
        self.idents = OrderedObj()

    def concat(self, other: "IdentifiersAcc") -> "IdentifiersAcc":
        return naive_object_concat(self, other)

    __add__ = concat

    def add_identifier(self, into_path: Path, ident: str, source: str) -> str:
        sources = Alt(
            sources
            for path_idents in self.idents.get(into_path.path)
            for sources in path_idents.get(ident)
        ).get_or(StringSet([source]))

        if source not in sources:
            raise ValueError(
                f"Cannot create ident {ident} in {into_path.path}, conflicts with existing definition."
            )

        self.idents += OrderedObj({into_path.path: OrderedObj({ident: StringSet([source])})})
        return ident

    def get_identifier_sources(self, path: Path, ident: str) -> Alt[StringSet]:
        return Alt(
            idents
            for path_idents in self.idents.get(path.path)
            for idents in path_idents.get(ident)
        )

    def unwrap_conflicts(self) -> List[Tuple[str, str, StringSet]]:
        return [
            (path, ident, sources)
            for path, mod in self.idents
            for ident, sources in mod
            if len(sources) > 1
        ]


class ImportsAcc:
    # into_path -> from_path -> from_ident -> into_idents
    imports: OrderedObj[OrderedObj[OrderedObj[StringSet]]]

    def __init__(self):
        self.imports = OrderedObj()

    def concat(self, other: "ImportsAcc") -> "ImportsAcc":
        return naive_object_concat(self, other)

    __add__ = concat

    def add_import(self, into_path: Path, from_path: ImportPath, from_ident: str, into_ident: str):
        self.imports += OrderedObj(
            {
                into_path.path: OrderedObj(
                    {from_path.rel_path: OrderedObj({from_ident: StringSet([into_ident])})}
                )
            }
        )

    def get_imported_as_idents(
        self, into_path: Path, from_path: ImportPath, from_ident: str
    ) -> Alt[StringSet]:
        return Alt(
            into_idents
            for imported_from in self.imports.get(into_path.path)
            for from_idents in imported_from.get(from_path.rel_path)
            for into_idents in from_idents.get(from_ident)
        )

    def render(self, into_path: str) -> List[str]:
        return Alt(
            [
                scripter.from_import(
                    ImportPath.as_python_module_path(rel_path),
                    *[
                        f"{from_ident} as {as_ident}" if from_ident != as_ident else from_ident
                        for from_ident, as_idents in decons
                        for as_ident in as_idents
                    ],
                )
                for rel_path, decons in imports
                if rel_path
            ]
            for imports in self.imports.get(into_path)
        ).get_or([])


class GeneratorAcc:
    idents: IdentifiersAcc
    imports: ImportsAcc
    content: OrderedObj[List]
    group_of_path: OrderedObj[StringSet]
    external_source_roots: Dict["GeneratorAcc", str]
    uniq: int

    def __init__(self):
        self.idents = IdentifiersAcc()
        self.imports = ImportsAcc()
        self.content = OrderedObj()
        self.group_of_path = OrderedObj()
        self.external_source_roots = {}
        self.uniq = 0

    def add_external_source_root(self, source_state: "GeneratorAcc", rel_path: str):
        self.external_source_roots[source_state] = rel_path

    def validate(self) -> "GeneratorAcc":
        path_errors = [
            f"Conflict in paths: Multiple ({' '.join(conflicts)}) types of {path} found"
            for path, path_groups in self.group_of_path
            for conflicts in Alt.unwrap_conflicts(path_groups)
        ]

        if path_errors:
            raise ValueError("\n".join(path_errors))

        conflicts = self.idents.unwrap_conflicts()
        if conflicts:
            raise ValueError(
                "Found conflicting identifiers:\n"
                + "\n  ".join(
                    f"ident {ident} was defined or imported into {path} by conflicting sources: {' '.join(sources)}"
                    for path, ident, sources in conflicts
                )
            )

        return self

    def render(self, comment_headers=Dict[str, List]) -> OrderedObj[str]:
        self.validate()

        return OrderedObj.from_iterable(
            OrderedObj(
                {
                    path: scripter.render(
                        scripter.comments(
                            lines
                            for group in groups
                            if group in comment_headers
                            for lines in comment_headers[group]
                        )
                        + self.imports.render(path)
                        + self.content.get(path).get_or([])
                    )
                }
            )
            for path, groups in self.group_of_path
            if self.content.get(path)
        )

    def add_content(self, path: Path, content: Union[str, List[str]]):
        if isinstance(content, str):
            content = [content]

        self.content += OrderedObj({path.path: content})

    def reserve_path(self, **lookup) -> Path:
        group, path = Alt(lookup.items()).unwrap()
        groups = self.group_of_path.get(path).get_or(StringSet([group]))

        if group in groups:
            self.group_of_path += OrderedObj({path: StringSet([group])})
            return Path(path)

        raise ValueError(
            f"Conflict:  cannot create file {path} for group {group}, already exists for {' '.join(groups)}"
        )

    def import_ident(
        self, into_path: Path, exported: Exported, as_ident: Optional[str] = None
    ) -> str:
        ident = exported.ident
        if as_ident is None:
            as_ident = ident

        # No imports actually required.
        if into_path == exported.path:
            return exported.ident

        export_path = exported.path

        # Handle pathing mapping of exports from external source states.
        if exported.source_state and exported.source_state is not self:
            rel_root = self.external_source_roots[exported.source_state]
            export_path = Path(os.path.join(rel_root, export_path.path))

        from_path = into_path.import_path_to(export_path)

        if not from_path.is_module and not exported.source_state.idents.get_identifier_sources(
            exported.path, ident
        ):
            raise ValueError(
                f"identifier {ident} required by {into_path} does not exist in {from_path}"
            )

        imported_as = self.imports.get_imported_as_idents(into_path, from_path, ident).get_or(
            StringSet([])
        )

        if imported_as:
            return sorted(imported_as)[0]

        as_ident = self.create_ident(into_path, as_ident, from_path.path.path)
        self.imports.add_import(into_path, from_path, ident, as_ident)

        return as_ident

    def create_ident(self, into_path: Path, as_ident: str, source: str) -> str:
        as_ident = get_safe_ident(as_ident)

        while source not in self.idents.get_identifier_sources(into_path, as_ident).get_or(
            StringSet([source])
        ):
            as_ident += "_"

        self.idents.add_identifier(into_path, as_ident, source)
        return as_ident

    def add_content_with_ident(
        self, path: Path, ident: str, scriptable: Callable[[str], Union[str, List]]
    ) -> str:
        self.idents.add_identifier(path, ident, self.get_unique_source(path))
        self.add_content(path, scriptable(ident))
        return ident

    def get_unique_source(self, path: Path) -> str:
        self.uniq += 1
        return path.path + "." + str(self.uniq)


Expression = Callable[[GeneratorAcc, Path], str]


class GeneratorContext:
    state: GeneratorAcc
    config: GeneratorConfig

    def __init__(self, state: GeneratorAcc, config: GeneratorConfig):
        self.state = state
        self.config = config


class AbstractGeneratorFileContext:
    path: Path
    state: GeneratorAcc

    def reserve_ident(self, ident: str) -> str:
        self.state.idents.add_identifier(self.path, ident, self.state.get_unique_source(self.path))
        return ident

    def export(self, ident: str, scriptable: Callable[[str], Union[str, List]]) -> Exported:
        assert self.state.idents.get_identifier_sources(self.path, ident).get_or(
            StringSet()
        ), "GeneratorFileContext.export called before identifier reserved."

        self.state.add_content(self.path, scriptable(ident))
        return Exported(self.path, ident, source_state=self.state)

    def import_ident(self, exported: Exported, as_ident: Optional[str] = None) -> str:
        return self.state.import_ident(self.path, exported, as_ident)

    def apply_expr(self, expression: Expression) -> str:
        return expression(self.state, self.path)


class GeneratorFileContext(AbstractGeneratorFileContext):
    parent: GeneratorContext

    def __init__(self, parent: GeneratorContext, path: Path):
        self.parent = parent
        self.path = path

    @property
    def state(self) -> GeneratorAcc:
        return self.parent.state

    @property
    def config(self) -> GeneratorConfig:
        return self.parent.config


class ExternFileContext(GeneratorFileContext):
    EXTERN_FILE: str = ""

    @cached_property
    def dumped_file(self) -> Path:
        content = open(self.EXTERN_FILE, encoding="utf-8").read()
        self.state.add_content(self.path, content)
        return self.path

    def export_extern(self, ident: str) -> Exported:
        return Exported(
            self.dumped_file,
            self.state.idents.add_identifier(self.dumped_file, ident, "extern"),
            source_state=self.state,
        )


def get_safe_ident(ident):
    while ident in KEYWORDS:
        ident += "_"
    return ident


KEYWORDS = {
    "False",
    "True",
    "class",
    "finally",
    "is",
    "return",
    "None",
    "continue",
    "for",
    "lambda",
    "try",
    "def",
    "from",
    "nonlocal",
    "while",
    "and",
    "del",
    "global",
    "not",
    "with",
    "as",
    "elif",
    "if",
    "or",
    "yield",
    "assert",
    "else",
    "import",
    "pass",
    "break",
    "except",
    "in",
    "raise",
}


def build(config: GeneratorConfig, output: OrderedObj[str]) -> Callable[[str], None]:
    build_env = BuildEnv(config.name, config.codegen_root)

    for path, contents in output:
        if contents:
            build_env.write_build_file(path, contents)

    return lambda output_dir: build_env.finalize(output_dir)



from typing import TypeVar, Generic, Tuple, List, Iterable, Any, Iterator, Mapping, Dict

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")

T = TypeVar("T")
R = TypeVar("R")


class StringSet(set):
    def add(self, element: str):
        super(StringSet, self).add(element)

    def __init__(self, items: Iterable[str] = None):
        super(StringSet, self).__init__(items or [])

    def __add__(self, other):
        return StringSet(self.union(other))

    concat = __add__


class OrderedObj(Generic[T]):
    obj: Mapping[str, T]
    ordering: List[str]

    def __init__(self, obj: Mapping[str, T] = None, ordering: List[str] = None):
        self.obj = obj or {}
        self.ordering = ordering or sorted(self.obj.keys())

    @classmethod
    def from_iterable(cls, items: Iterable["OrderedObj[T]"]) -> "OrderedObj[T]":
        result = OrderedObj()
        for o in items:
            result += o
        return result

    def __bool__(self):
        return bool(self.obj)

    def __len__(self):
        return len(self.obj)

    def concat(self, other: "OrderedObj[T]") -> "OrderedObj[T]":
        ordering = self.ordering + [i for i in other.ordering if i not in self.obj]
        obj: Dict[str, T] = {}

        for k in ordering:
            if k not in self.obj:
                obj[k] = other.obj[k]
            elif k not in other.obj:
                obj[k] = self.obj[k]
            else:
                left = self.obj[k]
                right = other.obj[k]
                obj[k] = left + right

        return OrderedObj(obj, ordering)

    __add__ = concat

    def keys(self) -> Iterable[str]:
        return self.ordering

    def values(self) -> Iterable[T]:
        for k in self.ordering:
            yield self.obj[k]

    def __iter__(self) -> Iterable[Tuple[str, T]]:
        for k in self.ordering:
            yield k, self.obj[k]

    def get(self, k: str) -> "Alt[T]":
        if k in self.obj:
            return Alt.lift(self.obj[k])
        return Alt.empty()


class Alt(Generic[A]):
    v: List[A]

    def __init__(self, v: Iterable[A]):
        if not isinstance(v, list):

            def takeone():
                for i in v:
                    yield i
                    break

            v = list(takeone())
        self.v = v

    @classmethod
    def lift(cls, v: A) -> "Alt[A]":
        return cls((v,))

    @classmethod
    def empty(cls) -> "Alt[Any]":
        return cls([])

    def unwrap(self) -> A:
        return self.v[0]

    def get_or(self, d: A) -> A:
        if not self:
            return d
        return self.unwrap()

    def __iter__(self) -> Iterable[A]:
        if self:
            yield self.unwrap()

    def concat(self, other: "Alt[A]") -> "Alt[A]":
        if self:
            return self
        return other

    def either(self, other: "Alt[A]") -> "Alt[A]":
        if self and other:
            raise ValueError("Unexpected conflict!")
        return self.concat(other)

    @staticmethod
    def unwrap_conflicts(items: List[A]) -> Iterator[List[A]]:
        if len(items) > 1:
            yield items

    __add__ = concat
    __xor__ = either

    def __bool__(self):
        return len(self.v) > 0


def naive_object_update(self, other: object) -> None:
    self_dict = self.__dict__
    for k, v in other.__dict__.items():
        if k in self_dict:
            self_dict[k] += v
        else:
            self_dict[k] = v


def naive_object_concat(self: A, other: A) -> A:
    result = self.__class__.__new__()
    result.__dict__ = dict(**self.__dict__)
    naive_object_update(result, other)
    return result



# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Struct, Primitive, Map, List
from typing import MutableMapping, MutableSequence
from ...schema.type import Type as ScaffoldType


class SchemaModule(Struct):
    """
    Metadata contained about a module.
    """

    module_name: str

    types_by_name: MutableMapping[str, ScaffoldType]

    types_dependency_ordering: MutableSequence[str]

    __field_constructors__ = [
        ("module_name", "module_name", Primitive.of(str), dict(optional=False),),
        (
            "types_by_name",
            "types_by_name",
            Map.of(ScaffoldType, dict(atleast_one=False)),
            dict(optional=False),
        ),
        (
            "types_dependency_ordering",
            "types_dependency_ordering",
            List.of(Primitive.of(str), dict(atleast_one=False)),
            dict(optional=False),
        ),
    ]



# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Enum


class SchemaPrimitiveType(Enum):
    INT = "int"
    DOUBLE = "double"
    STRING = "string"
    BOOL = "bool"
    ANY = "any"



# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Enum


class SchemaStructKind(Enum):
    """
    Scalars are non contained values
    Repeated are homogenous lists.  They can be upgraded or downgraded to and from Scalars.
    A repeated is indicated in a TypeDec by ending with '[]'
    Maps are homogenous 'dictionaries', whose key is always a string, mapping to js objects.
    A map is indicated in a TypeDec by ending with '{}'
    """

    SCALAR = "Scalar"
    REPEATED = "Repeated"
    MAP = "Map"



# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Struct, Map, List, Primitive
from typing import MutableMapping, Optional, MutableSequence
from ...schema.field import Field as ScaffoldField
from ...schema.type_struct import TypeStruct as ScaffoldTypeStruct
from ...schema.reference import Reference as ScaffoldReference


class SchemaType(Struct):
    """
    A wrapper type containing fields that can describe a Type, as well as its tag metadata.
    """

    # When this type is a struct, each of its fields and the type of that field is included
    # Exclusive with is_a and options
    fields: MutableMapping[str, ScaffoldField]

    # Set when this is type is an alias or simply an type expression (such as a generic).
    # Exclusive with having values for options or fields.
    is_a: Optional[ScaffoldTypeStruct]

    # The name and module information of this type's definition.
    named: ScaffoldReference

    # When this type is an enum includes the string values for each enum entry.  Note that each
    # target language may have different rules for the enum constant names, but these entries are
    # canonical resident values.
    # Exclusive with is_a and fields.
    options: MutableSequence[str]

    # General metadata given to a type.  Currently, atleast_one for Repeated types is supported.
    # Custom codegen can use these tags to implement semantic types on top of simple logic types.
    # In general, however, tags are considred optional and should not be required to
    # deserialize / serializeconsume correct logical values.
    tags: MutableSequence[str]

    __field_constructors__ = [
        (
            "fields",
            "fields",
            Map.of(ScaffoldField, dict(atleast_one=False)),
            dict(optional=False),
        ),
        ("is_a", "is_a", ScaffoldTypeStruct, dict(optional=True),),
        ("named", "named", ScaffoldReference, dict(optional=False),),
        (
            "options",
            "options",
            List.of(Primitive.of(str), dict(atleast_one=False)),
            dict(optional=False),
        ),
        (
            "tags",
            "tags",
            List.of(Primitive.of(str), dict(atleast_one=False)),
            dict(optional=False),
        ),
    ]



from idol.py.schema.struct_kind import StructKind
from .primitive_type import PrimitiveType
from ..codegen.schema.type_struct import SchemaTypeStruct


class TypeStruct(SchemaTypeStruct):
    @property
    def is_primitive(self):
        return not bool(self.reference.qualified_name)

    @property
    def is_reference(self):
        return not self.is_primitive

    @property
    def is_literal(self):
        return bool(self.literal)

    @property
    def is_alias(self):
        return bool(self.reference.type_name)

    @property
    def literal_value(self):
        if self.primitive_type == PrimitiveType.BOOL:
            return self.literal.bool
        elif self.primitive_type == PrimitiveType.DOUBLE:
            return self.literal.double
        elif self.primitive_type == PrimitiveType.INT:
            return self.literal.int
        elif self.primitive_type == PrimitiveType.STRING:
            return self.literal.string

    @property
    def type_display_name(self) -> str:
        result: str = self.primitive_type.value

        if self.struct_kind == StructKind.REPEATED:
            result += "[]"
        if self.struct_kind == StructKind.MAP:
            result += "{}"

        return result



# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Struct, Primitive


class SchemaReference(Struct):
    """
    A reference describes the location of a type in the module system.
    """

    # Just the module name
    module_name: str

    # The module_name.type_name string
    qualified_name: str

    # Just the type name
    type_name: str

    __field_constructors__ = [
        ("module_name", "module_name", Primitive.of(str), dict(optional=False),),
        ("qualified_name", "qualified_name", Primitive.of(str), dict(optional=False),),
        ("type_name", "type_name", Primitive.of(str), dict(optional=False),),
    ]



from os import getcwd

import argparse
import json
from typing import Dict, List, Union, MutableMapping
import sys

from idol.generator import GeneratorParams
from idol.functional import OrderedObj
from idol.__idol__ import Map
from idol.py.schema.module import Module


class CliConfig:
    flags: Dict[str, str]
    args: Dict[str, str]
    argparse: argparse.ArgumentParser

    def __init__(
        self,
        flags: Dict[str, str] = {},
        args: Dict[str, str] = {},
        description: str = "Codegenerator built on python",
    ):
        self.flags = flags
        self.args = args
        self.argparse = argparse.ArgumentParser(description=description)

        for arg, desc in args.items():
            self.argparse.add_argument("--" + arg, help=desc, nargs="*")

        for arg, desc in flags.items():
            self.argparse.add_argument("--" + arg, help=desc, action="store_true")

        self.argparse.add_argument(
            "-input_json",
            help="ignored when stdin is piped into this program, otherwise should be a json file containing the output of an idol run.",
        )


def start(config: CliConfig):
    args: argparse.Namespace = config.argparse.parse_args()
    if sys.stdin.isatty():
        data = open(args.input_json, "r").read()
    else:
        data = sys.stdin.read()

    return prepare_generator_params(vars(args), data)


def prepare_generator_params(
    options: Dict[str, Union[List[str], bool]], data: str
) -> GeneratorParams:
    modules = json.loads(data)
    Map.of(Module, {}).validate(modules)
    modules: MutableMapping[Module] = Map.of(Module, {}).wrap(modules)

    all_modules: OrderedObj[Module] = OrderedObj(modules)
    all_types = OrderedObj.from_iterable(m.types_as_ordered_obj() for m in all_modules.values())

    targets = options.get("target", [])
    scaffold_types = OrderedObj.from_iterable(
        modules[target].types_as_ordered_obj() for target in targets
    )

    return GeneratorParams(
        all_modules, all_types, scaffold_types, options.get("output", [getcwd()])[0], options
    )



from types import new_class

from collections.abc import MutableSequence
from typing import (
    TypeVar,
    MutableMapping,
    Any,
    Iterable,
    Tuple,
    Type,
    cast,
    Union,
    List as typingList,
    Dict,
    Iterator,
    Optional,
)
from enum import Enum as enumEnum


def with_metaclass(meta, *bases):
    return meta("NewBase", bases, {})


def get_list_scalar(value):
    while isinstance(value, list):
        if value:
            value = value[0]
        else:
            value = None
    return value


class IdolConstructor:
    @classmethod
    def validate(cls, json, path: typingList[str] = []):
        pass

    @classmethod
    def is_valid(cls, json) -> bool:
        try:
            cls.validate(json)
            return True
        except (ValueError, TypeError, KeyError):
            return False

    @classmethod
    def unwrap(cls, value) -> Any:
        return value

    @classmethod
    def wrap(cls, value) -> Any:
        return value

    @classmethod
    def expand(cls, json) -> Any:
        return json


class Primitive(IdolConstructor):
    type_constructor: Type

    def __new__(cls, *args, **kwargs):
        return cls.type_constructor(*args, **kwargs)

    @staticmethod
    def of(type_constructor: Type) -> Type["Primitive"]:
        cls = cast(
            Type["Primitive"], new_class(type_constructor.__name__, (Primitive,))
        )

        cls.type_constructor = type_constructor
        return cls

    @classmethod
    def expand(cls, json) -> Union[str, float, bool, int]:
        json = get_list_scalar(json)

        if json:
            return json

        if issubclass(cls.type_constructor, str):
            return ""
        if issubclass(cls.type_constructor, float):
            return 0.0
        if issubclass(cls.type_constructor, bool):
            return False
        if issubclass(cls.type_constructor, int):
            return 0

        raise TypeError("type_constructor was not one of str, float, bool, or int!")

    @classmethod
    def validate(cls, json, path: typingList[str] = []):
        if not issubclass(type(json), cls.type_constructor):
            raise TypeError(
                f"{'.'.join(path)} Expected type of {cls.type_constructor}, found {type(json)}"
            )


class Literal(IdolConstructor):
    value: Union[str, int, float, bool]

    @staticmethod
    def of(value: Union[str, int, float, bool]) -> Type["Literal"]:
        cls = cast(Type["Literal"], new_class(type(value).__name__, (Literal,)))
        cls.value = value
        return cls

    @classmethod
    def validate(cls, json, path=[]):
        if json != cls.value:
            raise ValueError(f"{'.'.join(path)} Expected to find literal {cls.value}")

    @classmethod
    def expand(cls, json) -> Union[str, int, float, bool]:
        json = get_list_scalar(json)

        if json is None or isinstance(json, type(cls.value)):
            json = cls.value

        return json


class Enum(IdolConstructor, enumEnum):
    @classmethod
    def validate(cls, json, path=[]):
        if not isinstance(json, str):
            raise TypeError(f"{'.'.join(path)} Expected a string, found {type(json)}")

        try:
            cls(json)
        except ValueError:
            raise ValueError(f"{'.'.join(path)} Value does not match enum type {cls}")

    @classmethod
    def expand(cls, json):
        json = get_list_scalar(json)

        if json is None:
            json = next(iter(cls)).value

        return json

    @classmethod
    def unwrap(cls, value: enumEnum):
        if isinstance(value, cls):
            return value.value
        return value

    @classmethod
    def wrap(cls: Type["Enum_T"], value) -> "Enum_T":
        return cls(value)


Enum_T = TypeVar("Enum_T", bound=Enum)


class List(IdolConstructor, MutableSequence):
    inner_constructor: Type[IdolConstructor]
    options: Dict[str, Any]

    @staticmethod
    def of(
        inner_constructor: Type[IdolConstructor],
        options: Optional[Dict[str, Any]] = None,
    ) -> Type["List"]:
        cls = cast(
            Type["List"], new_class(f"List[{inner_constructor.__name__}]", (List,))
        )
        cls.inner_constructor = inner_constructor
        cls.options = options
        return cls

    @classmethod
    def validate(cls, json, path=[]):
        if not isinstance(json, list):
            raise TypeError(f"{'.'.join(path)} Expected a list, found {type(json)}")

        if cls.options.get("atleast_one"):
            if not len(json):
                raise ValueError(
                    f"{'.'.join(path)} Expected at least one item, but it was empty"
                )

        for i, val in enumerate(json):
            cls.inner_constructor.validate(val, path + [str(i)])

    @classmethod
    def expand(cls, json) -> Any:
        if json is None:
            json = []

        if not isinstance(json, list):
            json = [json]

        if cls.options.get("atleast_one"):
            if not len(json):
                json.append(None)

        for i, val in enumerate(json):
            json[i] = cls.inner_constructor.expand(val)

        return json

    orig_list: typingList

    def __init__(self, orig_list: typingList):
        self.orig_list = orig_list

    def insert(self, index, v):
        return self.orig_list.insert(index, self.inner_constructor.unwrap(v))

    def __delitem__(self, i: int) -> None:
        return self.orig_list.__delitem__(i)

    def __len__(self) -> int:
        return len(self.orig_list)

    def __getitem__(self, i):
        return self.inner_constructor.wrap(self.orig_list.__getitem__(i))

    def __setitem__(self, key, value):
        self.orig_list.__setitem__(key, self.inner_constructor.unwrap(value))

    def __iter__(self) -> Iterator:
        for value in self.orig_list:
            yield self.inner_constructor.wrap(value)

    def __contains__(self, x: Any) -> bool:
        return self.inner_constructor.unwrap(x) in self.orig_list

    @classmethod
    def unwrap(cls, val):
        if isinstance(val, cls):
            return val.orig_list
        return val

    @classmethod
    def wrap(cls, val):
        return cls(val)


class Map(IdolConstructor, MutableMapping):
    inner_constructor: Type[IdolConstructor]
    options: Dict[str, Any]

    @staticmethod
    def of(
        inner_constructor: Type[IdolConstructor], options: Optional[Dict[str, Any]] = {}
    ) -> Type["Map"]:
        cls = cast(Type["Map"], new_class(f"Map[{inner_constructor.__name__}]", (Map,)))
        cls.inner_constructor = inner_constructor
        cls.options = options

        return cls

    @classmethod
    def validate(cls, json, path=[]):
        if not isinstance(json, dict):
            raise TypeError(f"{'.'.join(path)} Expected a dict, found {type(json)}")

        for key, val in json.items():
            cls.inner_constructor.validate(val, path + [key])

    @classmethod
    def expand(cls, json):
        json = get_list_scalar(json)

        if json is None:
            json = {}

        if not isinstance(json, dict):
            return json

        for key, val in json.items():
            json[key] = cls.inner_constructor.expand(val)

        return json

    orig_map: Dict[str, Any]

    def __init__(self, orig_map):
        self.orig_map = orig_map

    def __setitem__(self, k: str, v: Any):
        self.orig_map.__setitem__(k, self.inner_constructor.unwrap(v))

    def __delitem__(self, v: str):
        return self.orig_map.__delitem__(v)

    def __len__(self):
        return len(self.orig_map)

    def __iter__(self) -> Iterable[str]:
        for item in self.orig_map:
            yield item

    def __getitem__(self, i):
        return self.inner_constructor.wrap(self.orig_map.__getitem__(i))

    def items(self):
        for item in self.orig_map:
            yield (item, self[item])

    def values(self):
        for item in self.orig_map:
            yield self.inner_constructor.wrap(self[item])

    @classmethod
    def unwrap(cls, val):
        if isinstance(val, cls):
            return val.orig_map
        return val

    @classmethod
    def wrap(cls, val):
        return cls(val)


def create_struct_prop(attr, type: Type[IdolConstructor]):
    @property
    def prop(self):
        val = self.orig_data.get(attr, None)

        if val is None:
            return val

        return type.wrap(val)

    @prop.setter
    def prop(self, v):
        self.orig_data[attr] = type.unwrap(v)

    return prop


class StructMeta(type):
    def __new__(mcs: Type["Struct"], name, bases, dct):
        mcs = super().__new__(mcs, name, bases, dct)
        for field_name, prop_name, constructor, _ in getattr(
            mcs, "__field_constructors__", []
        ):
            setattr(mcs, prop_name, create_struct_prop(field_name, constructor))

        return mcs


class Struct(with_metaclass(StructMeta, IdolConstructor)):
    orig_data: Dict[str, Any]
    __field_constructors__: typingList[
        Tuple[str, str, Type[IdolConstructor], Dict[str, Any]]
    ] = []

    def __init__(self, orig_data: Dict[str, Any]):
        self.orig_data = orig_data

    def __str__(self):
        return str(self.orig_data)

    def __repr__(self):
        return repr(self.orig_data)

    @classmethod
    def validate(cls, json, path=[]):
        if not isinstance(json, dict):
            raise TypeError(f"{'.'.join(path)} Expected a dict, found {type(json)}")

        for field_name, prop_name, constructor, options in cls.__field_constructors__:
            val = json.get(field_name, None)
            optional = options.get("optional")

            if val is None:
                if optional:
                    continue
                else:
                    raise KeyError(
                        f"{'.'.join(path)} Missing required key {repr(field_name)}"
                    )

            constructor.validate(val, path + [field_name])

    @classmethod
    def expand(cls, json):
        json = get_list_scalar(json)

        if json is None:
            json = {}

        if not isinstance(json, dict):
            return json

        for field_name, prop_name, constructor, options in cls.__field_constructors__:
            val = json.get(field_name, None)
            optional = options.get("optional")

            if val is None:
                if optional:
                    json[field_name] = None
                    continue

            if optional:
                if not issubclass(constructor, List):
                    val = get_list_scalar(val)
                if val is not None:
                    val = constructor.expand(val)

                json[field_name] = val
            else:
                json[field_name] = constructor.expand(val)

        return json

    @classmethod
    def unwrap(cls, val):
        if isinstance(val, cls):
            return val.orig_data
        return val

    @classmethod
    def wrap(cls, val):
        return cls(val)



from setuptools import setup
import os.path

version = open(os.path.dirname(os.path.abspath(__file__)) + "/idol/VERSION").read()

setup(
    name="idol",
    version=version,
    description="idol codegen library tools",
    url="http://github.com/lyric-com/idol",
    author="Zach Collins",
    author_email="zach.collins@lyric.com",
    license="MIT",
    packages=["idol", "idol/py", "idol/py/codegen", "idol/py/codegen/schema", "idol/py/schema"],
    install_requires=["black", "cached_property"],
    python_require=">= 3.6",
    scripts=["idol/idol_py", "idol/idol_mar", "idol/idol_dataclass"],
    zip_safe=False,
    package_data={"": ["LICENSE", "idol_py", "idol_mar", "VERSION", "README"]},
)



import json
from typing import Iterable
from ..codegen.schema.type import SchemaType


class Type(SchemaType):
    @property
    def is_enum(self):
        return bool(len(self.options))

    def with_tags(self, tags: Iterable[str]) -> "Type":
        new = Type(json.loads(json.dumps(SchemaType.unwrap(self))))
        new.tags = list(tags)
        return new



from typing import Type, Callable, cast

import types

from marshmallow.fields import Function, Field


class Any(Function):
    def __init__(self, *args, **kwds):
        super(Any, self).__init__(lambda i: i, lambda i: i, *args, **kwds)


def wrap_field(field: Type[Field], name: str, *args, **kwds) -> Type[Field]:
    def wrapped_init(self, *a, **k):
        field.__init__(self, *list(arg() for arg in args), *a, **kwds, **k)

    new_class = types.new_class(
        name, (field,), exec_body=lambda class_dict: class_dict.update(__init__=wrapped_init)
    )

    return cast(Type[Field], new_class)



from typing import Iterable, List, Dict, Optional, Callable, Union

import black


def render(inner, mode=black.FileMode()):
    body = "\n".join(flatten_inner(inner))
    try:
        return black.format_str(body, mode=mode)
    except:
        print("")
        print(body)
        print("")
        raise


def from_import(module_name: str, *things) -> str:
    things_str = ", ".join(things)
    return f"from {module_name} import {things_str}"


def flatten_inner(inner, indent=0):
    if isinstance(inner, str):
        yield inner
        return

    try:
        for block in inner:
            for line in flatten_inner(block, indent + 1):
                yield ("    " * indent) + line
        return
    except TypeError:
        pass

    yield str(inner)


def comments(comments: Iterable[str]) -> List:
    return [f"# {s}" for s in comments for s in [s.replace("#", "\\#")]]


def assignment(ident: str, expr: str, typing=None) -> str:
    typing = f": {typing}" if typing else ""
    return f"{ident}{typing} = {expr}"


def assignable(expr: str, typing=None) -> Callable[[str], str]:
    return lambda ident: assignment(ident, expr, typing)


def typing(ident: str, typing: str) -> str:
    return f"{ident}: {typing}"


def shadow_assignment(ident: str, expr: str) -> List:
    return [
        assignment(ident, expr),
        assignment(index_access(invocation("locals"), literal(ident)), expr),
    ]


def commented(
    comment: Iterable[str], scriptable: Callable[[str], Union[List, str]]
) -> Callable[[str], List]:
    def wrapped(ident: str) -> List:
        script = scriptable(ident)
        return comments(comment) + (script if isinstance(script, list) else [script])

    return wrapped


def declare_and_shadow(
    declaration: str, shadow_expr: str, dec_typing: Optional[str] = None
) -> Callable[[str], List]:
    def scriptable(ident: str):
        return [assignment(ident, declaration, dec_typing)] + shadow_assignment(ident, shadow_expr)

    return scriptable


def invocation(callable: str, *args, **kwds):
    kwds = [f"{k}={v}" for k, v in kwds.items()]
    args = list(args) + kwds
    args = ", ".join(args)
    return f"{callable}({args})"


def index_access(expr: str, *keys):
    return f"{expr}[{', '.join(keys)}]"


def prop_access(value: str, *props):
    props = [value] + list(props)
    return ".".join(props)


def tuple(*values):
    return f"({', '.join(values)},)"


def type_parameterized(cons: str, *type_params) -> str:
    type_params = "".join(f"[{tp}]" for tp in type_params)
    return cons + type_params


def literal(value) -> str:
    return repr(value)


def thunk(expr: str) -> str:
    return f"(lambda: {expr})"


def array(values: Iterable[str]):
    return f"[{', '.join(values)}]"


def class_dec(
    class_name: str,
    super_classes: Iterable[str],
    body: Iterable,
    doc_str: Optional[Iterable[str]] = None,
    decorators: Optional[Iterable[str]] = None,
) -> List:
    super_str: str = ", ".join(super_classes) if super_classes else "object"
    return ["@" + dec for dec in (decorators or [])] + [
        f"class {class_name}({super_str}):",
        (['"""'] + list(doc_str) + ['"""'] if doc_str else []) + (list(body) or ["pass"]),
    ]


def nameable_class_dec(
    super_classes: Iterable[str],
    body: Iterable,
    doc_str: Optional[Iterable[str]] = None,
    decorators: Optional[Iterable[str]] = None,
) -> Callable[[str], List]:
    def nameable(ident: str) -> List:
        return class_dec(ident, super_classes, body, doc_str=doc_str, decorators=decorators)

    return nameable


def func_dec(
    func_name: str,
    args: List[str] = [],
    body: Iterable = ["pass"],
    kwds: Dict[str, str] = {},
    typing: str = "",
    decorators: Iterable[str] = [],
) -> List:
    if typing:
        typing = f" -> {typing}"

    return list(f"@{d}" for d in decorators) + [
        f"def {invocation(func_name, *args, **kwds)}{typing}:",
        list(body),
    ]


def ret(v: str):
    return f"return {v}"



#! /usr/bin/env python3

from cached_property import cached_property
from typing import List, Dict, Optional

from idol import scripter
from idol.cli import start, CliConfig
from idol.functional import OrderedObj, Alt
from idol.generator import (
    GeneratorParams,
    GeneratorConfig,
    build,
    GeneratorAcc,
    TypeDeconstructor,
    Path,
    get_material_type_deconstructor,
    TypeStructDeconstructor,
    ScalarDeconstructor,
    GeneratorContext,
    GeneratorFileContext,
    Exported,
    Expression,
    import_expr,
    get_safe_ident,
    get_tag_values,
    includes_tag,
    AbstractGeneratorFileContext,
    as_expression,
)
from idol.py.schema.primitive_type import PrimitiveType
from idol.py.schema.reference import Reference
from idol.py.schema.type import Type


class IdolDataCodegenStruct(GeneratorFileContext):
    codegen_file: "IdolDataCodegenFile"
    fields: OrderedObj["IdolDataCodegenTypeStruct"]

    def __init__(
        self, codegen_file: "IdolDataCodegenFile", fields: "OrderedObj[IdolDataCodegenTypeStruct]"
    ):
        self.fields = fields
        self.codegen_file = codegen_file
        super(IdolDataCodegenStruct, self).__init__(codegen_file.parent, codegen_file.path)
        self.reserve_ident(self.default_dataclass_name)

    @property
    def default_dataclass_name(self) -> str:
        return f"{self.codegen_file.t.named.as_qualified_ident}Dataclass"

    @cached_property
    def default_factory_expr(self) -> Alt[Expression]:
        return Alt(import_expr(dc) for dc in self.codegen_file.declared_field_constructor)

    @cached_property
    def declared_dataclass(self) -> Alt[Exported]:
        return Alt.lift(
            self.export(
                self.default_dataclass_name,
                scripter.nameable_class_dec(
                    [],
                    [
                        line
                        for field_name, field in self.fields
                        for typing_expr in field.typing_expr
                        for default_factory_expr in field.default_factory_expr
                        for line in (
                            scripter.comments(
                                get_tag_values(field.ts_decon.context.field_tags, "description")
                            )
                            + [
                                scripter.assignment(
                                    get_safe_ident(field_name),
                                    self.apply_expr(
                                        self.dataclass_field_expr(
                                            default_factory=default_factory_expr
                                        )
                                    ),
                                    typing=self.apply_expr(typing_expr),
                                )
                            ]
                        )
                    ],
                    doc_str=get_tag_values(self.codegen_file.t.tags, "description"),
                    decorators=[self.import_ident(Exported(Path("dataclasses"), "dataclass"))],
                ),
            )
        )

    def dataclass_field_expr(
        self, default: Optional[Expression] = None, default_factory: Optional[Expression] = None
    ) -> Optional[Expression]:
        def expr(state: GeneratorAcc, path: Path) -> str:
            kwds = {}
            if default:
                kwds["default"] = default(state, path)
            if default_factory:
                kwds["default_factory"] = default_factory(state, path)

            return scripter.invocation(
                state.import_ident(path, Exported(Path("dataclasses"), "field")), **kwds
            )

        return expr


class IdolDataCodegenEnum(GeneratorFileContext):
    codegen_file: "IdolDataCodegenFile"
    options: List[str]

    def __init__(self, codegen_file: "IdolDataCodegenFile", options: List[str]):
        self.options = options
        self.codegen_file = codegen_file
        super(IdolDataCodegenEnum, self).__init__(codegen_file.parent, codegen_file.path)

        self.reserve_ident(self.default_enum_name)

    @cached_property
    def default_factory_expr(self) -> Alt[Expression]:
        for declared_enum in self.codegen_file.declared_field_constructor:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.thunk(
                    scripter.invocation(
                        "next", scripter.invocation("iter", state.import_ident(path, declared_enum))
                    )
                )

            return Alt.lift(inner)

        return Alt.empty()

    @cached_property
    def declared_enum(self) -> Alt[Exported]:
        return Alt.lift(
            self.export(
                self.default_enum_name,
                scripter.nameable_class_dec(
                    [self.import_ident(Exported(Path("enum"), "Enum"))],
                    [
                        scripter.assignment(name.upper(), scripter.literal(name))
                        for name in self.options
                    ],
                    doc_str=get_tag_values(self.codegen_file.t.tags, "description"),
                ),
            )
        )

    @property
    def default_enum_name(self):
        return self.codegen_file.t.named.as_qualified_ident + "Enum"


class IdolDataCodegenScalar(GeneratorContext):
    idol_data: "IdolDataclass"
    scalar_dec: ScalarDeconstructor

    def __init__(self, parent: "IdolDataclass", scalar_decon: ScalarDeconstructor):
        super(IdolDataCodegenScalar, self).__init__(parent.state, parent.config)
        self.scalar_dec = scalar_decon
        self.idol_data = parent

    @cached_property
    def typing_expr(self) -> Alt[Expression]:
        return self.reference_import_expr ^ self.prim_typing_expr ^ self.literal_typing_expr

    @cached_property
    def default_factory_expr(self) -> Alt[Expression]:
        for type_decon in self.material_type_decon:
            codegen_file = self.idol_data.codegen_file(type_decon.t.named)
            return codegen_file.default_factory_expr

        return Alt(
            as_expression(scripter.thunk(scripter.literal(val)))
            for _, val in self.scalar_dec.get_literal()
        ) ^ Alt(
            as_expression(scripter.thunk(scripter.literal(self.scalar_primitive_default_map[prim])))
            for prim in self.scalar_dec.get_primitive()
            if prim in self.scalar_primitive_default_map
        )

    @cached_property
    def material_type_decon(self) -> Alt[TypeDeconstructor]:
        return Alt(
            get_material_type_deconstructor(
                self.idol_data.config.params.all_types,
                self.idol_data.config.params.all_types.obj[ref.qualified_name],
            )
            for ref in self.scalar_dec.get_alias()
        )

    @cached_property
    def reference_import_expr(self) -> Alt[Expression]:
        return Alt(
            import_expr(con)
            for ref in self.scalar_dec.get_alias()
            for con in self.idol_data.codegen_file(ref).declared_field_constructor
        )

    @cached_property
    def literal_typing_expr(self) -> Alt[Expression]:
        return Alt(
            self.scalar_primitive_type_map[prim_type]
            for prim_type, _ in self.scalar_dec.get_literal()
            if prim_type in self.scalar_primitive_type_map
        )

    @cached_property
    def prim_typing_expr(self) -> Alt[Expression]:
        scalar_prim: Alt[PrimitiveType] = self.scalar_dec.get_primitive()

        return Alt(
            self.scalar_primitive_type_map[prim_type]
            for prim_type in scalar_prim
            if prim_type in self.scalar_primitive_type_map
        )

    @cached_property
    def scalar_primitive_type_map(self) -> Dict[PrimitiveType, Expression]:
        return {
            PrimitiveType.BOOL: as_expression("bool"),
            PrimitiveType.INT: as_expression("int"),
            PrimitiveType.STRING: as_expression("str"),
            PrimitiveType.DOUBLE: as_expression("float"),
            PrimitiveType.ANY: import_expr(Exported(Path("typing"), "Any")),
        }

    @cached_property
    def scalar_primitive_default_map(self) -> Dict[PrimitiveType, any]:
        return {
            PrimitiveType.BOOL: False,
            PrimitiveType.INT: 0,
            PrimitiveType.STRING: "",
            PrimitiveType.DOUBLE: 0.0,
            PrimitiveType.ANY: {},
        }


class IdolDataCodegenTypeStruct(GeneratorContext):
    ts_decon: TypeStructDeconstructor
    idol_data: "IdolDataclass"

    IdolDataCodegenScalar = IdolDataCodegenScalar

    def __init__(self, parent: "IdolDataclass", ts_decon: TypeStructDeconstructor):
        super(IdolDataCodegenTypeStruct, self).__init__(parent.state, parent.config)
        self.ts_decon = ts_decon
        self.idol_data = parent

    @cached_property
    def inner_scalar(self) -> "Alt[IdolDataCodegenScalar]":
        return Alt(
            self.IdolDataCodegenScalar(self.idol_data, scalar_decon)
            for scalar_decon in (
                self.ts_decon.get_scalar() ^ self.ts_decon.get_map() ^ self.ts_decon.get_repeated()
            )
        )

    @cached_property
    def default_factory_expr(self) -> Alt[Expression]:
        if includes_tag(self.ts_decon.context.field_tags, "optional"):
            return Alt.lift(as_expression(scripter.thunk("None")))

        return (
            Alt(as_expression("list") for _ in self.ts_decon.get_repeated())
            ^ Alt(as_expression("dict") for _ in self.ts_decon.get_map())
            ^ Alt(
                expr
                for scalar in self.inner_scalar
                for expr in scalar.default_factory_expr
                if self.ts_decon.get_scalar()
            )
        )

    @cached_property
    def typing_expr(self) -> Alt[Expression]:
        def repeated_expr(scalar_typing_expr: Expression) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.index_access(
                    state.import_ident(path, Exported(Path("typing"), "List")),
                    scalar_typing_expr(state, path),
                )

            return inner

        def map_expr(scalar_typing_expr: Expression) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.index_access(
                    state.import_ident(path, Exported(Path("typing"), "Mapping")),
                    "str",
                    scalar_typing_expr(state, path),
                )

            return inner

        def optional_expr(inner_expr: Expression) -> Expression:
            def inner(state: GeneratorAcc, path: Path) -> str:
                return scripter.index_access(
                    state.import_ident(path, Exported(Path("typing"), "Optional")),
                    inner_expr(state, path),
                )

            return inner

        container_typing: Alt[Expression] = Alt(
            repeated_expr(scalar_typing_expr)
            for scalar in self.inner_scalar
            for scalar_typing_expr in scalar.typing_expr
            if self.ts_decon.get_repeated()
        )

        container_typing ^= Alt(
            map_expr(scalar_typing_expr)
            for scalar in self.inner_scalar
            for scalar_typing_expr in scalar.typing_expr
            if self.ts_decon.get_map()
        )

        typing = container_typing + Alt(
            expr for scalar in self.inner_scalar for expr in scalar.typing_expr
        )

        if includes_tag(self.ts_decon.context.field_tags, "optional"):
            return Alt(optional_expr(typing_expr) for typing_expr in typing)

        return typing


class IdolDataCodegenTypeStructDeclaration(IdolDataCodegenTypeStruct, AbstractGeneratorFileContext):
    path: Path
    codegen_file: "IdolDataCodegenFile"

    def __init__(self, parent: "IdolDataCodegenFile", ts_decon: TypeStructDeconstructor):
        super(IdolDataCodegenTypeStructDeclaration, self).__init__(parent.idol_data, ts_decon)
        self.codegen_file = parent
        self.path = parent.path

        self.reserve_ident(self.default_declared_typing_name)

    @cached_property
    def declared_field_typing(self) -> Alt[Exported]:
        return Alt(
            self.export(
                self.default_declared_typing_name, scripter.assignable(self.apply_expr(typing_expr))
            )
            for typing_expr in self.typing_expr
        )

    @cached_property
    def default_declared_typing_name(self) -> str:
        return self.codegen_file.t.named.type_name


class IdolDataScaffoldFile(GeneratorFileContext):
    t: Type
    idol_data: "IdolDataclass"

    def __init__(self, idol_data: "IdolDataclass", t: Type, path: Path):
        self.t = t
        self.idol_data = idol_data
        super(IdolDataScaffoldFile, self).__init__(idol_data, path)

        self.reserve_ident(self.default_dataclass_name)

    @cached_property
    def declared_dataclass(self) -> Alt[Exported]:
        type_decon = get_material_type_deconstructor(self.idol_data.config.params.all_types, self.t)

        codegen_ident: Alt[str] = Alt(
            self.import_ident(codegen_schema, self.default_dataclass_name + "Codegen")
            for codegen_schema in self.idol_data.codegen_file(self.t.named).declared_constructor
            if type_decon.get_struct() or type_decon.get_enum()
        )

        return Alt(
            self.export(
                self.default_dataclass_name,
                scripter.nameable_class_dec(
                    [ident], [], doc_str=get_tag_values(self.t.tags, "description")
                ),
            )
            for ident in codegen_ident
        )

    @property
    def default_dataclass_name(self) -> str:
        return self.t.named.type_name


class IdolDataCodegenFile(GeneratorFileContext):
    t: Type
    idol_data: "IdolDataclass"
    t_decon: TypeDeconstructor

    IdolDataCodegenTypeStructDeclaration = IdolDataCodegenTypeStructDeclaration
    IdolDataCodegenEnum = IdolDataCodegenEnum
    IdolDataCodegenStruct = IdolDataCodegenStruct
    IdolDataCodegenTypeStruct = IdolDataCodegenTypeStruct

    def __init__(self, idol_data: "IdolDataclass", t: Type, path: Path):
        self.t = t
        self.idol_data = idol_data
        self.t_decon = TypeDeconstructor(t)
        super(IdolDataCodegenFile, self).__init__(idol_data, path)

    @cached_property
    def default_factory_expr(self) -> Alt[Expression]:
        return (
            Alt(expr for struct in self.struct for expr in struct.default_factory_expr)
            ^ Alt(expr for enum in self.enum for expr in enum.default_factory_expr)
            ^ Alt(expr for ts in self.typestruct for expr in ts.default_factory_expr)
        )

    @cached_property
    def declared_constructor(self) -> Alt[Exported]:
        result: Alt[Exported] = Alt(
            v for typestruct in self.typestruct for v in typestruct.declared_field_typing
        )

        result ^= Alt(v for enum in self.enum for v in enum.declared_enum)
        result ^= Alt(v for struct in self.struct for v in struct.declared_dataclass)

        return result

    @cached_property
    def declared_field_constructor(self) -> Alt[Exported]:
        """
        Used as an indirection that may either be the declared_constructor generated by this
        codegen file, or the scaffold wrapper generated for types that are included in
        the scaffold.  Used by embedded references to this type.
        """
        return Alt(
            con
            for con in self.idol_data.scaffold_file(self.t.named).declared_dataclass
            if self.struct or self.enum
            if self.t.named.qualified_name in self.idol_data.config.params.scaffold_types.obj
        ) + self.declared_constructor

    @cached_property
    def typestruct(self) -> "Alt[IdolDataCodegenTypeStructDeclaration]":
        return Alt(
            self.IdolDataCodegenTypeStructDeclaration(self, ts_decon)
            for ts_decon in self.t_decon.get_typestruct()
        )

    @cached_property
    def enum(self) -> "Alt[IdolDataCodegenEnum]":
        return Alt(self.IdolDataCodegenEnum(self, options) for options in self.t_decon.get_enum())

    @cached_property
    def struct(self) -> "Alt[IdolDataCodegenStruct]":
        return Alt(
            self.IdolDataCodegenStruct(
                self,
                OrderedObj.from_iterable(
                    OrderedObj({k: self.IdolDataCodegenTypeStruct(self.idol_data, ts_decon)})
                    for k, ts_decon in fields
                ),
            )
            for fields in self.t_decon.get_struct()
        )


class IdolDataclass(GeneratorContext):
    scaffolds: Dict[str, "IdolDataScaffoldFile"]
    codegens: Dict[str, "IdolDataCodegenFile"]

    IdolDataScaffoldFile = IdolDataScaffoldFile
    IdolDataCodegenFile = IdolDataCodegenFile

    def __init__(self, config: GeneratorConfig):
        super(IdolDataclass, self).__init__(GeneratorAcc(), config)
        self.scaffolds = {}
        self.codegens = {}

    def scaffold_file(self, ref: Reference) -> "IdolDataScaffoldFile":
        path = self.state.reserve_path(**self.config.paths_of(scaffold=ref))
        t = self.config.params.all_types.obj[ref.qualified_name]

        if t.named.qualified_name not in self.scaffolds:
            self.scaffolds[t.named.qualified_name] = self.IdolDataScaffoldFile(self, t, path)

        return self.scaffolds[t.named.qualified_name]

    def codegen_file(self, ref: Reference) -> "IdolDataCodegenFile":
        path = self.state.reserve_path(**self.config.paths_of(codegen=ref))
        t = self.config.params.all_types.obj[ref.qualified_name]

        if t.named.qualified_name not in self.codegens:
            self.codegens[t.named.qualified_name] = self.IdolDataCodegenFile(self, t, path)

        return self.codegens[ref.qualified_name]

    def render(self) -> OrderedObj[str]:
        for i, t in enumerate(self.config.params.scaffold_types.values()):
            for export in self.scaffold_file(t.named).declared_dataclass:
                print(
                    f"Rendered {export.ident} to {export.path.path} ({i} / {len(self.config.params.scaffold_types)})"
                )
                break
            else:
                print(
                    f"Skipped {t.named.as_qualified_ident} ({i} / {len(self.config.params.scaffold_types)})"
                )

        return self.state.render(
            dict(
                codegen=[
                    "DO NOT EDIT",
                    "This file was generated by idol_data, any changes will be lost when idol_data is rerun again",
                ],
                scaffold=[
                    "This file was scaffold by idol_data, but it will not be overwritten, so feel free to edit.",
                    "This file will be regenerated if you delete it.",
                ],
            )
        )

    @classmethod
    def marshmallow_field_exported(cls, ident: str) -> Exported:
        return Exported(Path("marshmallow.fields"), ident)


def main():
    params: GeneratorParams = start(
        CliConfig(
            args={
                "target": "idol module names",
                "output": "a directory to generate the dataclasses into.",
            }
        )
    )

    config = GeneratorConfig(params)
    config.with_path_mappings(
        dict(
            codegen=config.in_codegen_dir(config.one_file_per_type),
            scaffold=config.one_file_per_type,
        )
    )

    idol_data = IdolDataclass(config)
    move_to = build(config, idol_data.render())
    move_to(params.output_dir)


if __name__ == "__main__":
    main()
